# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

__all__ = [ 'Moffat' ]

import numpy as np
import math

from . import _galsim
from .gsobject import GSObject
from .gsparams import GSParams
from .utilities import lazy_property, doc_inherit
from .errors import GalSimRangeError, GalSimIncompatibleValuesError, convert_cpp_errors


class Moffat(GSObject):
    r"""A class describing a Moffat surface brightness profile.

    The Moffat surface brightness profile is

    .. math::
        I(R) \sim \left(1 + (r/r_0)^2\right)^{-\beta}

    where :math:`r_0` is ``scale_radius``.

    The GalSim representation of a Moffat profile also includes an optional truncation beyond a
    given radius.

    For more information, refer to

        http://home.fnal.gov/~neilsen/notebook/astroPSF/astroPSF.html

    A Moffat can be initialized using one (and only one) of three possible size parameters:
    ``scale_radius``, ``fwhm``, or ``half_light_radius``.  Exactly one of these three is required.

    Parameters:
        beta:               The ``beta`` parameter of the profile.
        scale_radius:       The scale radius of the profile.  Typically given in arcsec.
                            [One of ``scale_radius``, ``fwhm``, or ``half_light_radius`` is
                            required.]
        half_light_radius:  The half-light radius of the profile.  Typically given in arcsec.
                            [One of ``scale_radius``, ``fwhm``, or ``half_light_radius`` is
                            required.]
        fwhm:               The full-width-half-max of the profile.  Typically given in arcsec.
                            [One of ``scale_radius``, ``fwhm``, or ``half_light_radius`` is
                            required.]
        trunc:              An optional truncation radius at which the profile is made to drop to
                            zero, in the same units as the size parameter.
                            [default: 0, indicating no truncation]
        flux:               The flux (in photons/cm^2/s) of the profile. [default: 1]
        gsparams:           An optional `GSParams` argument. [default: None]
    """
    _req_params = { "beta" : float }
    _opt_params = { "trunc" : float , "flux" : float }
    _single_params = [ { "scale_radius" : float, "half_light_radius" : float, "fwhm" : float } ]

    _is_axisymmetric = True
    _is_analytic_x = True
    _is_analytic_k = True

    # The conversion from hlr or fwhm to scale radius is complicated for Moffat, especially
    # since we allow it to be truncated, which matters for hlr.  So we do these calculations
    # in the C++-layer constructor.
    def __init__(self, beta, scale_radius=None, half_light_radius=None, fwhm=None, trunc=0.,
                 flux=1., gsparams=None):
        self._beta = float(beta)
        self._trunc = float(trunc)
        self._flux = float(flux)
        self._gsparams = GSParams.check(gsparams)

        if self._trunc == 0. and self._beta <= 1.1:
            raise GalSimRangeError("Moffat profiles with beta <= 1.1 must be truncated",
                                   beta, 1.1)
        if self._trunc < 0.:
            raise GalSimRangeError("Moffat trunc must be >= 0", self._trunc, 0.)

        # Parse the radius options
        if half_light_radius is not None:
            if scale_radius is not None or fwhm is not None:
                raise GalSimIncompatibleValuesError(
                    "Only one of scale_radius, half_light_radius, or fwhm may be specified",
                    half_light_radius=half_light_radius, scale_radius=scale_radius, fwhm=fwhm)
            self._hlr = float(half_light_radius)
            if self._trunc > 0. and self._trunc <= math.sqrt(2.) * self._hlr:
                raise GalSimRangeError("Moffat trunc must be > sqrt(2) * half_light_radius.",
                                       self._trunc, math.sqrt(2.) * self._hlr)
            with convert_cpp_errors():
                self._r0 = _galsim.MoffatCalculateSRFromHLR(self._hlr, self._trunc, self._beta)
            self._fwhm = 0.
        elif fwhm is not None:
            if scale_radius is not None:
                raise GalSimIncompatibleValuesError(
                    "Only one of scale_radius, half_light_radius, or fwhm may be specified",
                    half_light_radius=half_light_radius, scale_radius=scale_radius, fwhm=fwhm)
            self._fwhm = float(fwhm)
            self._r0 = self._fwhm / (2. * math.sqrt(2.**(1./self._beta) - 1.))
            self._hlr = 0.
        elif scale_radius is not None:
            self._r0 = float(scale_radius)
            self._hlr = 0.
            self._fwhm = 0.
        else:
            raise GalSimIncompatibleValuesError(
                "One of scale_radius, half_light_radius, or fwhm must be specified",
                half_light_radius=half_light_radius, scale_radius=scale_radius, fwhm=fwhm)

    @lazy_property
    def _sbp(self):
        with convert_cpp_errors():
            return _galsim.SBMoffat(self._beta, self._r0, self._trunc, self._flux,
                                    self.gsparams._gsp)

    @property
    def beta(self):
        """The beta parameter of this `Moffat` profile.
        """
        return self._beta
    @property
    def scale_radius(self):
        """The scale radius of this `Moffat` profile.
        """
        return self._r0
    @property
    def trunc(self):
        """The truncation radius (if any) of this `Moffat` profile.
        """
        return self._trunc

    @property
    def half_light_radius(self):
        """The half-light radius of this `Moffat` profile.
        """
        if self._hlr == 0.:
            self._hlr = self._sbp.getHalfLightRadius()
        return self._hlr

    @lazy_property
    def fwhm(self):
        """The FWHM of this `Moffat` profle.
        """
        if self._fwhm == 0.:
            self._fwhm = self._r0 * (2. * math.sqrt(2.**(1./self._beta) - 1.))
        return self._fwhm

    def __eq__(self, other):
        return (self is other or
                (isinstance(other, Moffat) and
                 self.beta == other.beta and
                 self.scale_radius == other.scale_radius and
                 self.trunc == other.trunc and
                 self.flux == other.flux and
                 self.gsparams == other.gsparams))

    def __hash__(self):
        return hash(("galsim.Moffat", self.beta, self.scale_radius, self.trunc, self.flux,
                     self.gsparams))

    def __repr__(self):
        return 'galsim.Moffat(beta=%r, scale_radius=%r, trunc=%r, flux=%r, gsparams=%r)'%(
            self.beta, self.scale_radius, self.trunc, self.flux, self.gsparams)

    def __str__(self):
        s = 'galsim.Moffat(beta=%s, scale_radius=%s'%(self.beta, self.scale_radius)
        if self.trunc != 0.:
            s += ', trunc=%s'%self.trunc
        if self.flux != 1.0:
            s += ', flux=%s'%self.flux
        s += ')'
        return s

    def __getstate__(self):
        d = self.__dict__.copy()
        d.pop('_sbp',None)
        return d

    def __setstate__(self, d):
        self.__dict__ = d

    @property
    def _maxk(self):
        return self._sbp.maxK()

    @property
    def _stepk(self):
        return self._sbp.stepK()

    @property
    def _has_hard_edges(self):
        return self._trunc != 0.

    @property
    def _max_sb(self):
        return self._sbp.maxSB()

    def _xValue(self, pos):
        return self._sbp.xValue(pos._p)

    def _kValue(self, kpos):
        return self._sbp.kValue(kpos._p)

    def _drawReal(self, image, jac=None, offset=(0.,0.), flux_scaling=1.):
        _jac = 0 if jac is None else jac.__array_interface__['data'][0]
        dx,dy = offset
        self._sbp.draw(image._image, image.scale, _jac, dx, dy, flux_scaling)

    def _shoot(self, photons, rng):
        self._sbp.shoot(photons._pa, rng._rng)

    def _drawKImage(self, image, jac=None):
        _jac = 0 if jac is None else jac.__array_interface__['data'][0]
        self._sbp.drawK(image._image, image.scale, _jac)

    @doc_inherit
    def withFlux(self, flux):
        return Moffat(beta=self.beta, scale_radius=self.scale_radius, trunc=self.trunc,
                      flux=flux, gsparams=self.gsparams)



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

__all__ = [ 'RandomKnots' ]

import numpy as np

from . import _galsim
from .gsparams import GSParams
from .gsobject import GSObject
from .position import PositionD
from .utilities import lazy_property, doc_inherit
from .errors import GalSimRangeError, GalSimValueError, GalSimIncompatibleValuesError
from .gaussian import Gaussian
from .random import BaseDeviate


class RandomKnots(GSObject):
    """
    A class for generating a set of point sources, following either a `Gaussian` profile or a
    specified input profile.

    Uses of this profile include representing an "irregular" galaxy, or
    adding this profile to an Exponential to represent knots of star formation.

    RandomKnots profiles have "shape noise" that depends on the number of point
    sources used.  For example, using the default Gaussian distribution, with
    100 points, the shape noise is g~0.05, and this will decrease as more
    points are added.  The profile can be sheared to give additional
    ellipticity, for example to follow that of an associated disk.

    The requested half light radius (hlr) should be thought of as a rough
    value.  With a finite number point sources the actual realized hlr will be
    noisy.

    .. note::

        If providing an input ``profile`` object, it must be "shoot-able".  Objects that
        cannot be drawn with ``method='phot'`` cannot be used as the ``profile`` parameter here.

    Parameters:
         npoints:           Number of point sources to generate.
         half_light_radius: Optional half light radius of the distribution of points.  This value
                            is used for a Gaussian distribution if an explicit profile is not sent.
                            This is the mean half light radius produced by an infinite number of
                            points.  A single instance will be noisy.  [default None]
         flux:              Optional total flux in all point sources.  This value is used for a
                            Gaussian distribution if an explicit profile is not sent. Defaults to
                            None if profile is sent, otherwise 1.  [default: None]
         profile:           Optional profile to use for drawing points.  If a profile is sent, the
                            half_light_radius and flux keywords are invalid.  [default: None]
         rng:               Optional random number generator. Can be any `galsim.BaseDeviate`.  If
                            None, the rng is created internally.  [default: None]
         gsparams:          Optional `GSParams` for the objects representing each point source.
                            [default: None]

    Attributes:
        npoints:                    The number of points to use as knots
        input_half_light_radius:    The input half_light_radius
        flux:                       The flux
        points:                     The array of x,y offsets used to create the point sources

    .. note::

        The algorithm was originally a modified version of that presented in
        https://arxiv.org/abs/1312.5514v3.  However, we now use the GalSim photon shooting
        mechanism, which allows the knots to trace any profile, not just a `Gaussian`.
    """
    # these allow use in a galsim configuration context

    _req_params = { "npoints" : int }
    _opt_params = {
        "flux" : float ,
        "half_light_radius": float,
        "profile": GSObject,
    }
    _takes_rng = True

    _has_hard_edges = False
    _is_axisymmetric = False
    _is_analytic_x = False
    _is_analytic_k = True

    def __init__(self, npoints, half_light_radius=None, flux=None, profile=None, rng=None,
                 gsparams=None):

        self._npoints=npoints
        self._half_light_radius = half_light_radius
        self._flux = flux
        self._profile=profile

        self._verify()
        self._gsparams = GSParams.check(gsparams)

        if rng is None:
            rng = BaseDeviate(rng)
            self._orig_rng=rng.duplicate()
        else:
            if not isinstance(rng, BaseDeviate):
                raise TypeError("rng must be an instance of galsim.BaseDeviate, got %s"%rng)
            self._orig_rng=rng.duplicate()
            # We won't use the rng yet, but make sure the original advances the right number
            # of values.
            rng.discard(2*npoints)

        if profile is None:
            if self._flux is None: self._flux = 1.0
            self._profile = Gaussian(half_light_radius=self._half_light_radius, flux=self._flux)

        else:
            self._flux=profile.flux
            try:
                # not all GSObjects have this attribute
                self._half_light_radius = profile.half_light_radius
            except Exception:
                self._half_light_radius = None


    @lazy_property
    def _sbp(self):
        fluxper = self._flux/self._npoints
        deltas = []
        for p in self.points:
            d = _galsim.SBDeltaFunction(fluxper, self.gsparams._gsp)
            d = _galsim.SBTransform(d, 0, p[0], p[1], 1.0, self.gsparams._gsp)
            deltas.append(d)
        return _galsim.SBAdd(deltas, self.gsparams._gsp)

    @property
    def input_half_light_radius(self):
        """
        Get the input half light radius (HLR).

        Note the input HLR is not necessarily the realized HLR,
        due to the finite number of points used in the profile.

        If a profile is sent, and that profile is a Transformation object (e.g.
        it has been sheared, its flux set, etc), then this value will be None.

        You can get the *calculated* half light radius using the calculateHLR
        method.  That value will be valid in all cases.
        """
        return self._half_light_radius

    @property
    def npoints(self):
        """The number of point sources.
        """
        return self._npoints

    @lazy_property
    def points(self):
        """A list of the locations (x,y) of the point sources.

        Technically, this is a numpy array of shape (npoints, 2).
        """
        rng = self._orig_rng.duplicate()
        photons = self._profile.shoot(self._npoints, rng)
        return np.column_stack([ photons.x, photons.y ])

    def calculateHLR(self):
        """
        calculate the half-light radius of the generated points
        """
        pts = self.points
        my,mx=pts.mean(axis=0)

        r=np.sqrt( (pts[:,0]-my)**2 + (pts[:,1]-mx)**2)

        hlr=np.median(r)

        return hlr

    def _verify(self):
        """
        type and range checking on the inputs
        """
        try:
            self._npoints = int(self._npoints)
        except ValueError as err:
            raise GalSimValueError("npoints should be a number: %s", str(err)) from None

        if self._npoints <= 0:
            raise GalSimRangeError("npoints must be > 0", self._npoints, 1)

        if self._profile is None:
            if self._half_light_radius is None:
                raise GalSimIncompatibleValuesError(
                    "half_light_radius required when not providing a profile")

            if self._half_light_radius <= 0.:
                raise GalSimRangeError(
                    "half_light_radius must be positive", self._half_light_radius, 0.)

        else:
            if self._flux is not None:
                raise GalSimIncompatibleValuesError("flux is invalid when providing a profile")

            if self._half_light_radius is not None:
                raise GalSimIncompatibleValuesError(
                    "half_light_radius is invalid when providing a profile")

            if not isinstance(self._profile, GSObject):
                raise GalSimIncompatibleValuesError("profile must be a GSObject")

    def __str__(self):
        rep = 'galsim.RandomKnots(%(npoints)d, profile=%(profile)s)'
        rep = rep % dict(
            npoints=self._npoints,
            profile=str(self._profile),
        )

        return rep

    def __repr__(self):
        rep = 'galsim.RandomKnots(%r, profile=%r, rng=%r, gsparams=%r)'%(
                self._npoints, self._profile, self._orig_rng, self._gsparams)
        return rep

    def __eq__(self, other):
        return (self is other or
                (isinstance(other, RandomKnots) and
                 self._npoints == other._npoints and
                 self._profile == other._profile and
                 self._orig_rng == other._orig_rng and
                 self._gsparams == other._gsparams))

    def __hash__(self):
        return hash(("galsim.RandomKnots", self._npoints, self._half_light_radius, self._flux,
                     self.gsparams))

    def __getstate__(self):
        d = self.__dict__.copy()
        d.pop('_sbp',None)
        return d

    def __setstate__(self, d):
        self.__dict__ = d

    @property
    def _maxk(self):
        return self._sbp.maxK()

    @property
    def _stepk(self):
        return self._sbp.stepK()

    @property
    def _centroid(self):
        return PositionD(self._sbp.centroid())

    @property
    def _positive_flux(self):
        return self._sbp.getPositiveFlux()

    @property
    def _negative_flux(self):
        return self._sbp.getNegativeFlux()

    @property
    def _max_sb(self):
        return self._sbp.maxSB()

    def _kValue(self, kpos):
        return self._sbp.kValue(kpos._p)

    def _shoot(self, photons, rng):
        self._sbp.shoot(photons._pa, rng._rng)

    def _drawKImage(self, image, jac=None):
        _jac = 0 if jac is None else jac.__array_interface__['data'][0]
        self._sbp.drawK(image._image, image.scale, _jac)

    # For all the transformations methods, apply to the internal profile, and remake points
    # in the correct locations.  This makes fft drawing much faster than the normal way
    # of applying the transformation to the k-space image.
    @doc_inherit
    def withFlux(self, flux):
        return RandomKnots(self.npoints, profile=self._profile.withFlux(flux),
                           rng=self._orig_rng.duplicate(), gsparams=self.gsparams)

    @doc_inherit
    def withScaledFlux(self, flux_ratio):
        if hasattr(flux_ratio, '__call__'):
            return GSObject.withScaledFlux(self, flux_ratio)
        else:
            return RandomKnots(self._npoints, profile=self._profile.withScaledFlux(flux_ratio),
                               rng=self._orig_rng.duplicate(), gsparams=self._gsparams)

    @doc_inherit
    def expand(self, scale):
        return RandomKnots(self._npoints, profile=self._profile.expand(scale),
                           rng=self._orig_rng.duplicate(), gsparams=self._gsparams)

    @doc_inherit
    def dilate(self, scale):
        return RandomKnots(self._npoints, profile=self._profile.dilate(scale),
                           rng=self._orig_rng.duplicate(), gsparams=self._gsparams)

    @doc_inherit
    def shear(self, *args, **kwargs):
        return RandomKnots(self._npoints, profile=self._profile.shear(*args, **kwargs),
                           rng=self._orig_rng.duplicate(), gsparams=self._gsparams)

    def _shear(self, shear):
        return RandomKnots(self._npoints, profile=self._profile._shear(shear),
                           rng=self._orig_rng.duplicate(), gsparams=self._gsparams)

    @doc_inherit
    def rotate(self, theta):
        return RandomKnots(self._npoints, profile=self._profile.rotate(theta),
                           rng=self._orig_rng.duplicate(), gsparams=self._gsparams)

    @doc_inherit
    def transform(self, dudx, dudy, dvdx, dvdy):
        return RandomKnots(self._npoints, profile=self._profile.transform(dudx,dudy,dvdx,dvdy),
                           rng=self._orig_rng.duplicate(), gsparams=self._gsparams)

    @doc_inherit
    def shift(self, *args, **kwargs):
        return RandomKnots(self._npoints, profile=self._profile.shift(*args, **kwargs),
                           rng=self._orig_rng.duplicate(), gsparams=self._gsparams)

    def _shift(self, dx, dy):
        return RandomKnots(self._npoints, profile=self._profile._shift(dx,dy),
                           rng=self._orig_rng.duplicate(), gsparams=self._gsparams)



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

__all__ = [ 'PhotonArray', 'PhotonOp', 'WavelengthSampler', 'FRatioAngles',
            'PhotonDCR', 'Refraction', 'FocusDepth',
            'PupilImageSampler', 'PupilAnnulusSampler', 'TimeSampler',
            'ScaleFlux', 'ScaleWavelength' ]

import numpy as np
import astropy.units as u

from . import _galsim
from .random import BaseDeviate
from .celestial import CelestialCoord
from ._utilities import lazy_property
from .angle import radians, arcsec, Angle, AngleUnit
from .errors import GalSimError, GalSimRangeError, GalSimValueError, GalSimUndefinedBoundsError
from .errors import GalSimIncompatibleValuesError, galsim_warn
from ._pyfits import pyfits
from . import dcr

# Add on more methods in the python layer

class PhotonArray:
    """The PhotonArray class encapsulates the concept of a collection of photons incident on
    a detector.

    A PhotonArray object is not typically constructed directly by the user.  Rather, it is
    typically constructed as the return value of the `GSObject.shoot` method.
    At this point, the photons only have x,y,flux values.

    Then there are a number of `Photon Operators`, which perform various modifications to the
    photons such as giving them wavelengths (`WavelengthSampler` or inclination angles
    (`FRatioAngles`) or move them around according to the effect of differential chromatic
    refraction (DCR; `PhotonDCR`).

    One could also add functionality to remove some photons due to fringing or vignetting,
    but these are not yet implemented.

    A PhotonArray instance has the following attributes, each of which is a numpy array:

    Attributes:
        x:          The incidence x position of the photons in image coordinates (pixels),
                    typically measured at the top of the detector.
        y:          The incidence y position of the photons in image coordinates (pixels),
                    typically measured at the top of the detector.
        flux:       The flux of the photons in units of photons. Typically, these are all 1,
                    but see the note below for reasons some photons might have flux != 1.
        dxdz:       The tangent of the inclination angles in the x direction.  Note that we define
                    the +z direction as towards towards the dielectric medium of the detector and
                    -z as towards vacuum; consequently, a photon with increasing x in time has
                    positive dxdz.
        dydz:       The tangent of the inclination angles in the y direction.  Note that we define
                    the +z direction as towards towards the dielectric medium of the detector and
                    -z as towards vacuum; consequently, a photon with increasing y in time has
                    positive dydz.
        wavelength  The wavelength of the photons (in nm)
        pupil_u:    Horizontal location of photon as it intersected the entrance pupil plane
                    (meters).
        pupil_v:    Vertical location of photon as it intersected the entrance pupil plane
                    (meters).
        time:       Time stamp for photon impacting the pupil plane (seconds).

    Unlike most GalSim objects (but like `Image`), PhotonArrays are mutable.  It is permissible
    to write values to the above attributes with code like::

        >>> photon_array.x += numpy.random.random(1000) * 0.01
        >>> photon_array.flux *= 20.
        >>> photon_array.wavelength = sed.sampleWavelength(photonarray.size(), bandpass)
        etc.

    All of these will update the existing numpy arrays being used by the photon_array instance.

    .. note::

        Normal photons have flux=1, but we allow for "fat" photons that combine the effect of
        several photons at once for efficiency.  Also, some profiles need to use negative flux
        photons to properly implement photon shooting (e.g. `InterpolatedImage`, which uses negative
        flux photons to get the interpolation correct).  Finally, when we "remove" photons, for
        better efficiency, we actually just set the flux to 0 rather than recreate new numpy arrays.

    The initialization constructs a PhotonArray to hold N photons, but does not set the values of
    anything yet.  The constructor allocates space for the x,y,flux arrays, since those are always
    needed.  The other arrays are only allocated on demand if the user accesses these attributes.

    Parameters:
        N:          The number of photons to store in this PhotonArray.  This value cannot be
                    changed.
        x:          Optionally, the initial x values. [default: None]
        y:          Optionally, the initial y values. [default: None]
        flux:       Optionally, the initial flux values. [default: None]
        dxdz:       Optionally, the initial dxdz values. [default: None]
        dydz:       Optionally, the initial dydz values. [default: None]
        wavelength: Optionally, the initial wavelength values (in nm). [default: None]
        pupil_u:    Optionally, the initial pupil_u values. [default: None]
        pupil_v:    Optionally, the initial pupil_v values. [default: None]
        time:       Optionally, the initial time values. [default: None]
    """
    def __init__(
        self, N, x=None, y=None, flux=None, dxdz=None, dydz=None, wavelength=None,
        pupil_u=None, pupil_v=None, time=None
    ):
        # Only x, y, flux are built by default, since these are always required.
        # The others we leave as None unless/until they are needed.
        self._x = np.zeros(N, dtype=float)
        self._y = np.zeros(N, dtype=float)
        self._flux = np.zeros(N, dtype=float)
        self._dxdz = None
        self._dydz = None
        self._wave = None
        self._pupil_u = None
        self._pupil_v = None
        self._time = None
        self._is_corr = False

        # These give reasonable errors if x,y,flux are the wrong size/type
        if x is not None: self.x = x
        if y is not None: self.y = y
        if flux is not None: self.flux = flux
        if dxdz is not None: self.dxdz = dxdz
        if dydz is not None: self.dydz = dydz
        if wavelength is not None: self.wavelength = wavelength
        if pupil_u is not None: self.pupil_u = pupil_u
        if pupil_v is not None: self.pupil_v = pupil_v
        if time is not None: self.time = time

    @classmethod
    def fromArrays(
        cls, x, y, flux, dxdz=None, dydz=None, wavelength=None, pupil_u=None, pupil_v=None,
        time=None, is_corr=False,
    ):
        """Create a PhotonArray from pre-allocated numpy arrays without any copying.

        The normal PhotonArray constructor always allocates new arrays and copies any provided
        initial values into those new arrays.  This class method, by constrast, constructs a
        PhotonArray that references existing numpy arrays, so that any PhotonOps or photon shooting
        of GSObjects applied to the resulting PhotonArray will also be reflected in the original
        arrays.

        Note that the input arrays must all be the same length, have dtype float64 and be
        c_contiguous.

        Parameters:
            x:          X values.
            y:          X values.
            flux:       Flux values.
            dxdz:       Optionally, the initial dxdz values. [default: None]
            dydz:       Optionally, the initial dydz values. [default: None]
            wavelength: Optionally, the initial wavelength values (in nm). [default: None]
            pupil_u:    Optionally, the initial pupil_u values (in m). [default: None]
            pupil_v:    Optionally, the initial pupil_v values (in m). [default: None]
            time:       Optionally, the initial time values (in s). [default: None]
            is_corr:    Whether or not the photons are correlated. [default: False]
        """
        args = [x, y, flux]
        argnames = ['x', 'y', 'flux']
        for a, aname in zip(
            [dxdz, dydz, wavelength, pupil_u, pupil_v, time],
            ['dxdz', 'dydz', 'wavelength', 'pupil_u', 'pupil_v', 'time']
        ):
            if a is not None:  # don't check optional args that are None
                args.append(a)
                argnames.append(aname)

        N = len(x)
        for a, aname in zip(args, argnames):
            if not isinstance(a, np.ndarray):
                raise TypeError("Argument {} must be an ndarray".format(aname))
            if not a.dtype == np.float64:
                raise TypeError("Array {} dtype must be np.float64".format(aname))
            if not len(a) == N:
                raise ValueError("Arrays must all be the same length")
            if not a.flags.c_contiguous:
                raise ValueError("Array {} must be c_contiguous".format(aname))

        return cls._fromArrays(x, y, flux, dxdz, dydz, wavelength, pupil_u, pupil_v, time, is_corr)

    @classmethod
    def _fromArrays(
        cls, x, y, flux, dxdz=None, dydz=None, wavelength=None, pupil_u=None, pupil_v=None,
        time=None, is_corr=False
    ):
        """Same as `fromArrays`, but no sanity checking of inputs.
        """
        ret = PhotonArray.__new__(PhotonArray)
        ret._x = x
        ret._y = y
        ret._flux = flux
        ret._dxdz = dxdz
        ret._dydz = dydz
        ret._wave = wavelength
        ret._pupil_u = pupil_u
        ret._pupil_v = pupil_v
        ret._time = time
        ret._is_corr = False
        if is_corr:
            from .deprecated import depr
            depr('is_corr=True', 2.5, '',
                "We don't think this is necessary anymore.  If you have a use case that "
                "requires it, please open an issue.")
            ret._is_corr = is_corr
        return ret

    @classmethod
    def concatenate(cls, photon_arrays):
        """Create a single PhotonArray from a list of multiple PhotonArrays.

        The size of the created PhotonArray will be the sum of the sizes of the given arrays,
        and the values will be concatenations of the values in each.

        .. note::
            The optional value arrays (e.g. dxdz, wavelength, etc.) must be given in
            all the given photon_arrays or in none of them.  This is not checked.

        Parameters:
            photon_arrays:  A list of PhotonArray objects to be concatenated.
        """
        p1 = photon_arrays[0]
        kwargs = {
            'x': np.concatenate([p.x for p in photon_arrays]),
            'y': np.concatenate([p.y for p in photon_arrays]),
            'flux': np.concatenate([p.flux for p in photon_arrays]),
        }

        if p1.hasAllocatedAngles():
            kwargs['dxdz'] = np.concatenate([p.dxdz for p in photon_arrays])
            kwargs['dydz'] = np.concatenate([p.dydz for p in photon_arrays])
        if p1.hasAllocatedWavelengths():
            kwargs['wavelength'] = np.concatenate([p.wavelength for p in photon_arrays])
        if p1.hasAllocatedPupil():
            kwargs['pupil_u'] = np.concatenate([p.pupil_u for p in photon_arrays])
            kwargs['pupil_v'] = np.concatenate([p.pupil_v for p in photon_arrays])
        if p1.hasAllocatedTimes():
            kwargs['time'] = np.concatenate([p.time for p in photon_arrays])

        return cls._fromArrays(**kwargs)

    def size(self):
        """Return the size of the photon array.  Equivalent to ``len(self)``.
        """
        return len(self._x)

    def __len__(self):
        return len(self._x)

    @property
    def x(self):
        """The incidence x position in image coordinates (pixels), typically at the top of
        the detector.
        """
        return self._x
    @x.setter
    def x(self, value):
        self._x[:] = value

    @property
    def y(self):
        """The incidence y position in image coordinates (pixels), typically at the top of
        the detector.
        """
        return self._y
    @y.setter
    def y(self, value):
        self._y[:] = value

    @property
    def flux(self):
        """The flux of the photons.
        """
        return self._flux
    @flux.setter
    def flux(self, value):
        self._flux[:] = value

    @property
    def dxdz(self):
        """The tangent of the inclination angles in the x direction: dx/dz.
        """
        if not self.hasAllocatedAngles():
            from .deprecated import depr
            depr('dxdz accessed before being set.', 2.5,
                 'Angle arrays should be set or explicitly allocated before being accessed.',
                 'For now, accessing dxdz allocates an array with all zeros. '
                 'This will become an error in a future version (probably 3.0).')
            self.allocateAngles()
        return self._dxdz
    @dxdz.setter
    def dxdz(self, value):
        self.allocateAngles()
        self._dxdz[:] = value

    @property
    def dydz(self):
        """The tangent of the inclination angles in the y direction: dy/dz.
        """
        if not self.hasAllocatedAngles():
            from .deprecated import depr
            depr('dydz accessed before being set.', 2.5,
                 'Angle arrays should be set or explicitly allocated before being accessed.',
                 'For now, accessing dydz allocates an array with all zeros. '
                 'This will become an error in a future version (probably 3.0).')
            self.allocateAngles()
        return self._dydz
    @dydz.setter
    def dydz(self, value):
        self.allocateAngles()
        self._dydz[:] = value

    @property
    def wavelength(self):
        """The wavelength of the photons (in nm).
        """
        if not self.hasAllocatedWavelengths():
            from .deprecated import depr
            depr('wavelength accessed before being set.', 2.5,
                 'Wavelength array should be set or explicitly allocated before being accessed.',
                 'For now, accessing wavelength allocates arrays with all zeros. '
                 'This will become an error in a future version (probably 3.0).')
            self.allocateWavelengths()
        return self._wave
    @wavelength.setter
    def wavelength(self, value):
        self.allocateWavelengths()
        self._wave[:] = value

    @property
    def pupil_u(self):
        """Horizontal location of photon as it intersected the entrance pupil plane.
        """
        if not self.hasAllocatedPupil():
            from .deprecated import depr
            depr('pupil_u accessed before being set.', 2.5,
                 'Pupil arrays should be set or explicitly allocated before being accessed.',
                 'For now, accessing pupil_u allocates arrays with all zeros. '
                 'This will become an error in a future version (probably 3.0).')
            self.allocatePupil()
        return self._pupil_u
    @pupil_u.setter
    def pupil_u(self, value):
        self.allocatePupil()
        self._pupil_u[:] = value

    @property
    def pupil_v(self):
        """Vertical location of photon as it intersected the entrance pupil plane.
        """
        if not self.hasAllocatedPupil():
            from .deprecated import depr
            depr('pupil_v accessed before being set.', 2.5,
                 'Pupil arrays should be set or explicitly allocated before being accessed.',
                 'For now, accessing pupil_v allocates arrays with all zeros. '
                 'This will become an error in a future version (probably 3.0).')
            self.allocatePupil()
        return self._pupil_v
    @pupil_v.setter
    def pupil_v(self, value):
        self.allocatePupil()
        self._pupil_v[:] = value

    @property
    def time(self):
        """Time stamp of when photon encounters the pupil plane.
        """
        if not self.hasAllocatedTimes():
            from .deprecated import depr
            depr('time accessed before being set.', 2.5,
                 'Time array should be set or explicitly allocated before being accessed.',
                 'For now, accessing time allocates arrays with all zeros. '
                 'This will become an error in a future version (probably 3.0).')
            self.allocateTimes()
        return self._time
    @time.setter
    def time(self, value):
        self.allocateTimes()
        self._time[:] = value

    def hasAllocatedAngles(self):
        """Returns whether the arrays for the incidence angles `dxdz` and `dydz` have been
        allocated.
        """
        return self._dxdz is not None and self._dydz is not None

    def allocateAngles(self):
        """Allocate memory for the incidence angles, `dxdz` and `dydz`.
        """
        if self._dxdz is None:
            self._dxdz = np.zeros_like(self._x)
            self._dydz = np.zeros_like(self._x)
            self.__dict__.pop('_pa', None)

    def hasAllocatedWavelengths(self):
        """Returns whether the `wavelength` array has been allocated.
        """
        return self._wave is not None

    def allocateWavelengths(self):
        """Allocate the memory for the `wavelength` array.
        """
        if self._wave is None:
            self._wave = np.zeros_like(self._x)
            self.__dict__.pop('_pa', None)

    def hasAllocatedPupil(self):
        """Returns whether the arrays for the pupil coordinates `pupil_u` and `pupil_v` have been
        allocated.
        """
        return self._pupil_u is not None and self._pupil_v is not None

    def allocatePupil(self):
        """Allocate the memory for the pupil coordinates, `pupil_u` and `pupil_v`.
        """
        if self._pupil_u is None:
            self._pupil_u = np.zeros_like(self._x)
            self._pupil_v = np.zeros_like(self._x)

    def hasAllocatedTimes(self):
        """Returns whether the array for the time stamps `time` has been allocated.
        """
        return self._time is not None

    def allocateTimes(self):
        """Allocate the memory for the time stamps, `time`.
        """
        if self._time is None:
            self._time = np.zeros_like(self._x)

    def isCorrelated(self):
        """Returns whether the photons are correlated
        """
        from .deprecated import depr
        depr('isCorrelated', 2.5, '',
                "We don't think this is necessary anymore.  If you have a use case that "
                "requires it, please open an issue.")
        return self._is_corr

    def setCorrelated(self, is_corr=True):
        """Set whether the photons are correlated
        """
        from .deprecated import depr
        depr('setCorrelated', 2.5, '',
                "We don't think this is necessary anymore.  If you have a use case that "
                "requires it, please open an issue.")
        self._is_corr = is_corr
        self.__dict__.pop('_pa', None)

    def getTotalFlux(self):
        """Return the total flux of all the photons.
        """
        return self.flux.sum()

    def setTotalFlux(self, flux):
        """Rescale the photon fluxes to achieve the given total flux.

        Parameter:
            flux:       The target flux
        """
        self.scaleFlux(flux / self.getTotalFlux())

    def scaleFlux(self, scale):
        """Rescale the photon fluxes by the given factor.

        Parameter:
            scale:      The factor by which to scale the fluxes.
        """
        self._flux *= scale

    def scaleXY(self, scale):
        """Scale the photon positions (`x` and `y`) by the given factor.

        Parameter:
            scale:      The factor by which to scale the positions.
        """
        self._x *= scale
        self._y *= scale

    def assignAt(self, istart, rhs):
        """Assign the contents of another `PhotonArray` to this one starting at istart.

        Parameters:
            istart:     The first index at which to insert new values.
            rhs:        The other `PhotonArray` from which to get the values.
        """
        from .deprecated import depr
        depr("PhotonArray.assignAt", 2.5, "copyFrom(rhs, slice(istart, istart+rhs.size()))")
        if istart + rhs.size() > self.size():
            raise GalSimValueError(
                "The given rhs does not fit into this array starting at %d"%istart, rhs)
        s = slice(istart, istart + rhs.size())
        self._copyFrom(rhs, s, slice(None))

    def copyFrom(self, rhs, target_indices=slice(None), source_indices=slice(None),
                 do_xy=True, do_flux=True, do_other=True):
        """Copy some contents of another `PhotonArray` to some elements of this one.

        Specifically each element of rhs[source_indices] is mapped to self[target_indices].
        The values s1 and s2 may be slices, list of indices, or anything else that is a valid
        key for a numpy array.

        Parameters:
            rhs:            The `PhotonArray` from which to get values.
            target_indices: The indices at which to assign values in the current PhotonArray (self).
                            [default: slice(None)]
            source_indices: The indices from which to get values from the PhotonArray, ``rhs``.
                            [default: slice(None)]
            do_xy:          Whether to include copying the x and y arrays. [default: True]
            do_flux:        Whether to include copying the flux array. [default: True]
            do_other:       Whether to include copying the other arrays (angles, wavelength,
                            pupil positions, time). [default: True]
        """
        try:
            a1 = self.flux[target_indices]
            # Numpy is flexible about allowing slices outside the range of the array.
            # Rather than try to check all possible ways the indices can be invalid, we
            # just make sure that at least some elements come back from the numpy call.
            n1 = len(np.atleast_1d(a1))
            assert n1 > 0
        except (IndexError, AssertionError):
            raise GalSimValueError("target_indices is invalid for the target PhotonArray",
                                   target_indices)
        try:
            a2 = rhs.flux[source_indices]
            n2 = len(np.atleast_1d(a2))
            assert n2 > 0
        except (IndexError, AssertionError) as e:
            raise GalSimValueError("source_indices is invalid for the source PhotonArray",
                                   source_indices)
        if n1 != n2:
            raise GalSimIncompatibleValuesError(
                "target_indices and source_indices do not reference the same number of elements"
                "in their respective PhotonArrays ({} and {} respectively)".format(n1, n2),
                dict(target_indices=target_indices, source_indices=source_indices))

        self._copyFrom(rhs, target_indices, source_indices, do_xy, do_flux, do_other)

    def _copyFrom(self, rhs, target_indices, source_indices, do_xy=True, do_flux=True,
                  do_other=True):
        """Equivalent to self.copyFrom(rhs, target_indices, source_indices), but without any
        checks that the indices are valid.
        """
        # Aliases for notational convenience.
        s1 = target_indices
        s2 = source_indices

        if do_xy:
            self.x[s1] = rhs.x[s2]
            self.y[s1] = rhs.y[s2]
        if do_flux:
            self.flux[s1] = rhs.flux[s2]
        if do_other and rhs.hasAllocatedAngles():
            self.allocateAngles()
            self.dxdz[s1] = rhs.dxdz[s2]
            self.dydz[s1] = rhs.dydz[s2]
        if do_other and rhs.hasAllocatedWavelengths():
            self.allocateWavelengths()
            self.wavelength[s1] = rhs.wavelength[s2]
        if do_other and rhs.hasAllocatedPupil():
            self.allocatePupil()
            self.pupil_u[s1] = rhs.pupil_u[s2]
            self.pupil_v[s1] = rhs.pupil_v[s2]
        if do_other and rhs.hasAllocatedTimes():
            self.allocateTimes()
            self.time[s1] = rhs.time[s2]

    def convolve(self, rhs, rng=None):
        """Convolve this `PhotonArray` with another.

        ..note::

            If both self and rhs have wavelengths, angles, pupil coordinates or times assigned,
            then the values from the first array (i.e. self) take precedence.
        """
        if rhs.size() != self.size():
            raise GalSimIncompatibleValuesError("PhotonArray.convolve with unequal size arrays",
                                                self_pa=self, rhs=rhs)
        if rhs.hasAllocatedAngles() and not self.hasAllocatedAngles():
            self.dxdz = rhs.dxdz
            self.dydz = rhs.dydz

        if rhs.hasAllocatedWavelengths() and not self.hasAllocatedWavelengths():
            self.wavelength = rhs.wavelength

        if rhs.hasAllocatedPupil() and not self.hasAllocatedPupil():
            self.pupil_u = rhs.pupil_u
            self.pupil_v = rhs.pupil_v

        if rhs.hasAllocatedTimes() and not self.hasAllocatedTimes():
            self.time = rhs.time

        rng = BaseDeviate(rng)
        self._pa.convolve(rhs._pa, rng._rng)

    def __repr__(self):
        s = "galsim.PhotonArray(%d, x=array(%r), y=array(%r), flux=array(%r)"%(
                self.size(), self.x.tolist(), self.y.tolist(), self.flux.tolist())
        if self.hasAllocatedAngles():
            s += ", dxdz=array(%r), dydz=array(%r)"%(self.dxdz.tolist(), self.dydz.tolist())
        if self.hasAllocatedWavelengths():
            s += ", wavelength=array(%r)"%(self.wavelength.tolist())
        if self.hasAllocatedPupil():
            s += ", pupil_u=array(%r), pupil_v=array(%r)"%(self.pupil_u.tolist(), self.pupil_v.tolist())
        if self.hasAllocatedTimes():
            s += ", time=array(%r)"%(self.time.tolist())
        s += ")"
        return s

    def __str__(self):
        return "galsim.PhotonArray(%d)"%self.size()

    def __getstate__(self):
        d = self.__dict__.copy()
        d.pop('_pa',None)
        return d

    def __setstate__(self, d):
        self.__dict__ = d

    __hash__ = None

    def __eq__(self, other):
        return (self is other or
                (isinstance(other, PhotonArray) and
                 np.array_equal(self.x,other.x) and
                 np.array_equal(self.y,other.y) and
                 np.array_equal(self.flux,other.flux) and
                 self.hasAllocatedAngles() == other.hasAllocatedAngles() and
                 self.hasAllocatedWavelengths() == other.hasAllocatedWavelengths() and
                 self.hasAllocatedPupil() == other.hasAllocatedPupil() and
                 self.hasAllocatedTimes() == other.hasAllocatedTimes() and
                 (np.array_equal(self.dxdz,other.dxdz) if self.hasAllocatedAngles() else True) and
                 (np.array_equal(self.dydz,other.dydz) if self.hasAllocatedAngles() else True) and
                 (np.array_equal(self.wavelength,other.wavelength)
                    if self.hasAllocatedWavelengths() else True) and
                 (np.array_equal(self.pupil_u,other.pupil_u)
                    if self.hasAllocatedPupil() else True) and
                 (np.array_equal(self.pupil_v,other.pupil_v)
                    if self.hasAllocatedPupil() else True) and
                 (np.array_equal(self.time,other.time)
                    if self.hasAllocatedTimes() else True)
                ))

    def __ne__(self, other):
        return not self == other

    @lazy_property
    def _pa(self):
        #assert(self._x.strides[0] == self._x.itemsize)
        #assert(self._y.strides[0] == self._y.itemsize)
        #assert(self._flux.strides[0] == self._flux.itemsize)
        _x = self._x.__array_interface__['data'][0]
        _y = self._y.__array_interface__['data'][0]
        _flux = self._flux.__array_interface__['data'][0]
        _dxdz = _dydz = _wave = 0
        if self.hasAllocatedAngles():
            #assert(self._dxdz.strides[0] == self._dxdz.itemsize)
            #assert(self._dydz.strides[0] == self._dydz.itemsize)
            _dxdz = self._dxdz.__array_interface__['data'][0]
            _dydz = self._dydz.__array_interface__['data'][0]
        if self.hasAllocatedWavelengths():
            #assert(self._wave.strides[0] == self._wave.itemsize)
            _wave = self._wave.__array_interface__['data'][0]
        return _galsim.PhotonArray(int(self.size()), _x, _y, _flux, _dxdz, _dydz, _wave,
                                   self._is_corr)

    def addTo(self, image):
        """Add flux of photons to an image by binning into pixels.

        Photons in this `PhotonArray` are binned into the pixels of the input
        `Image` and their flux summed into the pixels.  The `Image` is assumed to represent
        surface brightness, so photons' fluxes are divided by image pixel area.
        Photons past the edges of the image are discarded.

        Parameters:
            image:      The `Image` to which the photons' flux will be added.

        Returns:
            the total flux of photons the landed inside the image bounds.
        """
        if not image.bounds.isDefined():
            raise GalSimUndefinedBoundsError(
                "Attempting to PhotonArray::addTo an Image with undefined Bounds")
        return self._pa.addTo(image._image)

    @classmethod
    def makeFromImage(cls, image, max_flux=1., rng=None):
        """Turn an existing `Image` into a `PhotonArray` that would accumulate into this image.

        The flux in each non-zero pixel will be turned into 1 or more photons with random positions
        within the pixel bounds.  The ``max_flux`` parameter (which defaults to 1) sets an upper
        limit for the absolute value of the flux of any photon.  Pixels with abs values > maxFlux
        will spawn multiple photons.

        Parameters:
            image:      The image to turn into a `PhotonArray`
            max_flux:   The maximum flux value to use for any output photon [default: 1]
            rng:        A `BaseDeviate` to use for the random number generation [default: None]

        Returns:
            a `PhotonArray`
        """
        # TODO: This corresponds to the Nearest interpolant.  It would be worth figuring out how
        #       to implement other (presumably better) interpolation options here.

        max_flux = float(max_flux)
        if (max_flux <= 0):
            raise GalSimRangeError("max_flux must be positive", max_flux, 0.)
        total_flux = np.abs(image.array).sum(dtype=float)

        # This goes a bit over what we actually need, but not by much.  Worth it to not have to
        # worry about array reallocations.
        N = int(np.prod(image.array.shape) + total_flux / max_flux)
        photons = cls(N)

        rng = BaseDeviate(rng)
        N = photons._pa.setFrom(image._image, max_flux, rng._rng)
        photons._x = photons.x[:N]
        photons._y = photons.y[:N]
        photons._flux = photons.flux[:N]

        if image.scale != 1. and image.scale is not None:
            photons.scaleXY(image.scale)
        return photons

    def write(self, file_name):
        """Write a `PhotonArray` to a FITS file.

        The output file will be a FITS binary table with a row for each photon in the `PhotonArray`.
        Columns will include 'id' (sequential from 1 to nphotons), 'x', 'y', and 'flux'.
        Additionally, the columns 'dxdz', 'dydz', and 'wavelength' will be included if they are
        set for this `PhotonArray` object.

        The file can be read back in with the classmethod `PhotonArray.read`::

            >>> photons.write('photons.fits')
            >>> photons2 = galsim.PhotonArray.read('photons.fits')

        Parameters:
            file_name:  The file name of the output FITS file.
        """
        cols = []
        cols.append(pyfits.Column(name='id', format='J', array=range(self.size())))
        cols.append(pyfits.Column(name='x', format='D', array=self.x))
        cols.append(pyfits.Column(name='y', format='D', array=self.y))
        cols.append(pyfits.Column(name='flux', format='D', array=self.flux))

        if self.hasAllocatedAngles():
            cols.append(pyfits.Column(name='dxdz', format='D', array=self.dxdz))
            cols.append(pyfits.Column(name='dydz', format='D', array=self.dydz))

        if self.hasAllocatedWavelengths():
            cols.append(pyfits.Column(name='wavelength', format='D', array=self.wavelength))

        if self.hasAllocatedPupil():
            cols.append(pyfits.Column(name='pupil_u', format='D', array=self.pupil_u))
            cols.append(pyfits.Column(name='pupil_v', format='D', array=self.pupil_v))

        if self.hasAllocatedTimes():
            cols.append(pyfits.Column(name='time', format='D', array=self.time))

        cols = pyfits.ColDefs(cols)
        table = pyfits.BinTableHDU.from_columns(cols)
        fits.writeFile(file_name, table)

    @classmethod
    def read(cls, file_name):
        """Create a `PhotonArray`, reading the photon data from a FITS file.

        The file being read in is not arbitrary.  It is expected to be a file that was written
        out with the `PhotonArray.write` method.::

            >>> photons.write('photons.fits')
            >>> photons2 = galsim.PhotonArray.read('photons.fits')

        Parameters:
            file_name:  The file name of the input FITS file.
        """
        with pyfits.open(file_name) as fits:
            data = fits[1].data
        N = len(data)
        names = data.columns.names

        photons = cls(N, x=data['x'], y=data['y'], flux=data['flux'])
        if 'dxdz' in names:
            photons.dxdz = data['dxdz']
            photons.dydz = data['dydz']
        if 'wavelength' in names:
            photons.wavelength = data['wavelength']
        if 'pupil_u' in names:
            photons.pupil_u = data['pupil_u']
            photons.pupil_v = data['pupil_v']
        if 'time' in names:
            photons.time = data['time']
        return photons


class PhotonOp:
    """A base class for photon operators, which just defines the interface.

    Photon operators are designed to apply some physical effect to a bundle of photons.  They
    may adjust the fluxes in some way, or the positions, maybe in a wavelength-dependent way, etc.
    They are typically applied via a ``photon_ops`` argument to the `GSObject.drawImage` method.
    The order typically matters, so the operators are applied in the order they appear in the list.
    """
    def applyTo(self, photon_array, local_wcs=None, rng=None):
        """Apply the photon operator to a PhotonArray.

        Parameters:
            photon_array:   A `PhotonArray` to apply the operator to.
            local_wcs:      A `LocalWCS` instance defining the local WCS for the current photon
                            bundle in case the operator needs this information.  [default: None]
            rng:            A random number generator to use if needed. [default: None]
        """
        raise NotImplementedError("Cannot call applyTo on a pure PhotonOp object")

    # These simpler versions of == and hash are fine.
    def __eq__(self, other):
        return repr(self) == repr(other)

    def __hash__(self):
        return hash(repr(self))


class WavelengthSampler(PhotonOp):
    """A photon operator that uses sed.sampleWavelength to set the wavelengths array of a
    `PhotonArray`.

    Parameters:
        sed:        The `SED` to use for the objects spectral energy distribution.
        bandpass:   A `Bandpass` object representing a filter, or None to sample over the full
                    `SED` wavelength range.
        rng:        If provided, a random number generator that is any kind of `BaseDeviate`
                    object. If ``rng`` is None, one will be automatically created, using the
                    time as a seed. [default: None]
        npoints:    Number of points `DistDeviate` should use for its internal interpolation
                    tables. [default: None, which uses the `DistDeviate` default]
    """
    _opt_params = { 'npoints' : int }

    def __init__(self, sed, bandpass=None, rng=None, npoints=None):
        if rng is not None:
            from .deprecated import depr
            depr('WavelengthSampler(..., rng)', 2.3, '',
                 'Instead provide rng when calling applyTo, drawImage, etc.')
        self.sed = sed
        self.bandpass = bandpass
        self.rng = rng
        self.npoints = npoints

    def applyTo(self, photon_array, local_wcs=None, rng=None):
        """Assign wavelengths to the photons sampled from the SED * Bandpass.

        Parameters:
            photon_array:   A `PhotonArray` to apply the operator to.
            local_wcs:      A `LocalWCS` instance defining the local WCS for the current photon
                            bundle in case the operator needs this information.  [default: None]
            rng:            A random number generator to use if needed. [default: None]
        """
        rng = rng if rng is not None else self.rng
        if photon_array.hasAllocatedWavelengths():
            galsim_warn("Wavelengths already set before applying WavelengthSampler. "
                        "This is most likely an error.")
        photon_array.wavelength = self.sed.sampleWavelength(
                photon_array.size(), self.bandpass, rng=rng, npoints=self.npoints)

    def __str__(self):
        return "galsim.WavelengthSampler(sed=%s, bandpass=%s, rng=%s, npoints=%s)"%(
            self.sed, self.bandpass, self.rng, self.npoints)

    def __repr__(self):
        return "galsim.WavelengthSampler(sed=%r, bandpass=%r, rng=%r, npoints=%r)"%(
            self.sed, self.bandpass, self.rng, self.npoints)


class FRatioAngles(PhotonOp):
    """A photon operator that assigns photon directions based on the f/ratio and obscuration.

    Assigns arrival directions at the focal plane for photons, drawing from a uniform
    brightness distribution between the obscuration angle and the edge of the pupil defined
    by the f/ratio of the telescope.  The angles are expressed in terms of slopes dx/dz
    and dy/dz.

    Parameters:
        fratio:         The f/ratio of the telescope (e.g. 1.2 for LSST)
        obscuration:    Linear dimension of central obscuration as fraction of aperture
                        linear dimension. [0., 1.).  [default: 0.0]
        rng:            A random number generator to use or None, in which case an rng
                        will be automatically constructed for you. [default: None]
    """
    _req_params = { 'fratio' : float }
    _opt_params = { 'obscuration' : float }

    def __init__(self, fratio, obscuration=0.0, rng=None):
        if fratio < 0:
            raise GalSimRangeError("The f-ratio must be positive.", fratio, 0.)
        if obscuration < 0 or obscuration >= 1:
            raise GalSimRangeError("Invalid obscuration.", obscuration, 0., 1.)
        if rng is not None:
            from .deprecated import depr
            depr('FRatioAngles(..., rng)', 2.3, '',
                 'Instead provide rng when calling applyTo, drawImage, etc.')

        self.fratio = fratio
        self.obscuration = obscuration
        self.rng = rng


    def applyTo(self, photon_array, local_wcs=None, rng=None):
        """Assign directions to the photons in photon_array.

        Parameters:
            photon_array:   A `PhotonArray` to apply the operator to.
            local_wcs:      A `LocalWCS` instance defining the local WCS for the current photon
                            bundle in case the operator needs this information.  [default: None]
            rng:            A random number generator to use if needed. [default: None]
        """
        gen = BaseDeviate(rng).as_numpy_generator()

        n_photons = len(photon_array)

        # The f/ratio is the ratio of the focal length to the diameter of the aperture of
        # the telescope.  The angular radius of the field of view is defined by the
        # ratio of the radius of the aperture to the focal length
        pupil_angle = np.arctan(0.5 / self.fratio)  # radians
        obscuration_angle = np.arctan(0.5 * self.obscuration / self.fratio)

        # Generate azimuthal angles for the photons
        phi = gen.uniform(0, 2*np.pi, size=n_photons)

        # Generate inclination angles for the photons, which are uniform in sin(theta) between
        # the sine of the obscuration angle and the sine of the pupil radius
        sintheta = gen.uniform(np.sin(obscuration_angle), np.sin(pupil_angle), size=n_photons)

        # Assign the directions to the arrays. In this class the convention for the
        # zero of phi does not matter but it would if the obscuration is dependent on
        # phi
        tantheta = np.sqrt(np.square(sintheta) / (1. - np.square(sintheta)))
        photon_array.dxdz = tantheta * np.sin(phi)
        photon_array.dydz = tantheta * np.cos(phi)

    def __str__(self):
        return "galsim.FRatioAngles(fratio=%s, obscration=%s, rng=%s)"%(
            self.fratio, self.obscuration, self.rng)

    def __repr__(self):
        return "galsim.FRatioAngles(fratio=%r, obscration=%r, rng=%r)"%(
            self.fratio, self.obscuration, self.rng)


class PhotonDCR(PhotonOp):
    r"""A photon operator that applies the effect of differential chromatic refraction (DCR)
    and optionally the chromatic dilation due to atmospheric seeing.

    Due to DCR, blue photons land closer to the zenith than red photons.  Kolmogorov turbulence
    also predicts that blue photons get spread out more by the atmosphere than red photons,
    specifically FWHM is proportional to :math:`\lambda^{-0.2}`.  Both of these effects can be
    implemented by wavelength-dependent shifts of the photons.

    Since DCR depends on the zenith angle and the parallactic angle (which is the position angle of
    the zenith measured from North through East) of the object being drawn, these must be specified
    via keywords.  There are four ways to specify these values:

    1) explicitly provide ``zenith_angle`` as a keyword of type `Angle`, and ``parallactic_angle``
       will be assumed to be 0 by default.
    2) explicitly provide both ``zenith_angle`` and ``parallactic_angle`` as keywords of type
       `Angle`.
    3) provide the coordinates of the object ``obj_coord`` and the coordinates of the zenith
       ``zenith_coord`` as keywords of type `CelestialCoord`.
    4) provide the coordinates of the object ``obj_coord`` as a `CelestialCoord`, the hour angle
       of the object ``HA`` as an `Angle`, and the latitude of the observer ``latitude`` as an
       `Angle`.

    DCR also depends on temperature, pressure and water vapor pressure of the atmosphere.  The
    default values for these are expected to be appropriate for LSST at Cerro Pachon, Chile, but
    they are broadly reasonable for most observatories.

    This photon op is intended to match the functionality of `ChromaticAtmosphere`, but acting
    on the photon array rather than as a `ChromaticObject`.  The photons will need to have
    wavelengths defined in order to work.

    .. warning::
        The alpha parameter is only appropriate for stars.  This photon op will act on
        all of the photons, so applying a chromatic dilation according to the chromatic
        seeing is the wrong thing to do when the surface brightness being rendered is
        not a pure PSF.  As such, the default is alpha=0, not -0.2, which would be
        appropriate for Kolmogorov turbulence.

    Parameters:
        base_wavelength:    Wavelength (in nm) represented by the fiducial photon positions
        scale_unit:         Units used for the positions of the photons.  [default: galsim.arcsec]
        alpha:              Power law index for wavelength-dependent seeing.  This should only
                            be used if doing a star-only simulation.  It is not correct when
                            drawing galaxies. [default: 0.]
        zenith_angle:       `Angle` from object to zenith, expressed as an `Angle`. [default: 0]
        parallactic_angle:  Parallactic angle, i.e. the position angle of the zenith, measured
                            from North through East.  [default: 0]
        obj_coord:          Celestial coordinates of the object being drawn as a `CelestialCoord`.
                            [default: None]
        zenith_coord:       Celestial coordinates of the zenith as a `CelestialCoord`.
                            [default: None]
        HA:                 Hour angle of the object as an `Angle`. [default: None]
        latitude:           Latitude of the observer as an `Angle`. [default: None]
        pressure:           Air pressure, either as an astropy Quantity or a float in units of
                            kiloPascals.  [default: 69.328 kPa]
        temperature:        Temperature, either as an astropy Quantity or a float in units of
                            Kelvin.  [default: 293.15 K]
        H2O_pressure:       Water vapor pressure, either as an astropy Quantity or a float in units
                            of kiloPascals.  [default: 1.067 kPa]
    """
    _req_params = { 'base_wavelength' : float }
    _opt_params = { 'scale_unit' : str,
                    'alpha' : float,
                    'parallactic_angle' : Angle,
                    'latitude' : Angle,
                    'pressure' : (float, u.Quantity),
                    'temperature' : (float, u.Quantity),
                    'H2O_pressure' : (float, u.Quantity)
                  }
    _single_params = [ { 'zenith_angle' : Angle, 'HA' : Angle, 'zenith_coord' : CelestialCoord } ]

    def __init__(self, base_wavelength, scale_unit=arcsec, **kwargs):
        # This matches the code in ChromaticAtmosphere.
        self.base_wavelength = base_wavelength

        if isinstance(scale_unit, str):
            scale_unit = AngleUnit.from_name(scale_unit)
        self.scale_unit = scale_unit
        self.alpha = kwargs.pop('alpha', 0.)

        self.zenith_angle, self.parallactic_angle, self.kw = dcr.parse_dcr_angles(**kwargs)
        # Convert any weather data to the appropriate units
        p = self.kw.get('pressure', None)
        if p is not None and isinstance(p, u.Quantity):
            self.kw['pressure'] = p.to_value(u.kPa)
        t = self.kw.get('temperature', None)
        if t is not None and isinstance(t, u.Quantity):
            self.kw['temperature'] = t.to_value(u.K)
        h = self.kw.get('H2O_pressure', None)
        if h is not None and isinstance(h, u.Quantity):
            self.kw['H2O_pressure'] = h.to_value(u.kPa)

        # Any remaining kwargs will get forwarded to galsim.dcr.get_refraction
        # Check that they're valid
        for kw in self.kw:
            if kw not in ('temperature', 'pressure', 'H2O_pressure'):
                raise TypeError("Got unexpected keyword: {0}".format(kw))

        self.base_refraction = dcr.get_refraction(self.base_wavelength, self.zenith_angle,
                                                  **self.kw)

    def applyTo(self, photon_array, local_wcs=None, rng=None):
        """Apply the DCR effect to the photons

        Parameters:
            photon_array:   A `PhotonArray` to apply the operator to.
            local_wcs:      A `LocalWCS` instance defining the local WCS for the current photon
                            bundle in case the operator needs this information.  [default: None]
            rng:            A random number generator to use if needed. [default: None]
        """
        if not photon_array.hasAllocatedWavelengths():
            raise GalSimError("PhotonDCR requires that wavelengths be set")
        if local_wcs is None:
            raise TypeError("PhotonDCR requires a local_wcs to be provided to applyTo")

        w = photon_array.wavelength
        cenx = local_wcs.origin.x
        ceny = local_wcs.origin.y

        # Apply the wavelength-dependent scaling
        if self.alpha != 0.:
            scale = (w/self.base_wavelength)**self.alpha
            photon_array.x = scale * (photon_array.x - cenx) + cenx
            photon_array.y = scale * (photon_array.y - ceny) + ceny

        # Apply DCR
        shift_magnitude = dcr.get_refraction(w, self.zenith_angle, **self.kw)
        shift_magnitude -= self.base_refraction
        shift_magnitude *= radians / self.scale_unit
        sinp, cosp = self.parallactic_angle.sincos()

        du = -shift_magnitude * sinp
        dv = shift_magnitude * cosp

        dx = local_wcs._x(du, dv)
        dy = local_wcs._y(du, dv)
        photon_array.x += dx
        photon_array.y += dy

    def __repr__(self):
        s = "galsim.PhotonDCR(base_wavelength=%r, scale_unit=%r, alpha=%r, "%(
                self.base_wavelength, self.scale_unit, self.alpha)
        s += "zenith_angle=%r, parallactic_angle=%r"%(self.zenith_angle, self.parallactic_angle)
        for k in sorted(self.kw):
            s += ", %s=%r"%(k, self.kw[k])
        s += ")"
        return s


class Refraction(PhotonOp):
    """A photon operator that refracts photons (manipulating their dxdz and dydz values) at an
    interface, commonly the interface between vacuum and silicon at the surface of a CCD.

    Assumes that the surface normal is along the z-axis.  If the refraction would result in total
    internal reflection, then those photon's dxdz and dydz values are set to NaN, and flux values
    set to 0.0.

    Parameters:
        index_ratio:    The ratio of the refractive index on the far side of the interface to the
                        near side.  Can be given as a number or a callable function.  In the
                        latter case, the function should accept a numpy array of vacuum wavelengths
                        as input and return a numpy array of refractive index ratios.
    """
    _req_params = { 'index_ratio' : float }

    def __init__(self, index_ratio):
        self.index_ratio = index_ratio

    def applyTo(self, photon_array, local_wcs=None, rng=None):
        """Refract photons

        Parameters:
            photon_array:   A `PhotonArray` to apply the operator to.
            local_wcs:      A `LocalWCS` instance defining the local WCS for the current photon
                            bundle in case the operator needs this information.  [default: None]
            rng:            A random number generator to use if needed. [default: None]
        """
        if hasattr(self.index_ratio, '__call__'):
            index_ratio = self.index_ratio(photon_array.wavelength)
        else:
            index_ratio = self.index_ratio
        # Here's the math avoiding any actual trig function calls:
        #
        #        x1 = dr/dz
        #        ------+
        #         \    |
        #          \   |
        #           \  |  dz/dz = 1
        #            \ |
        #             \|         n1
        #  ------------------------
        #              |\        n2
        #              |\
        # dz'/dz' = 1  | \
        #              | \
        #              |  \
        #              +---
        #              x2 = dr'/dz'
        #
        # Solve Snell's law for x2 as fn of x1:
        #   n1 sin(th1) = n2 sin(th2)
        #   n1 x1 / sqrt(1 + x1^2) = n2 x2 / sqrt(1 + x2^2)
        #   n1^2 x1^2 (1 + x2^2) = n2^2 x2^2 (1 + x1^2)
        #   n1^2 x1^2 = x2^2 (n2^2 (1 + x1^2) - n1^2 x1^2)
        #   x1^2 = x2^2 ((n2/n1)^2 (1 + x1^2) - x1^2)
        #   x1 = x2 sqrt( (n2/n1)^2 (1 + x1^2) - x1^2 )
        #      = x2 sqrt( (n2/n1)^2 (1 + x1^2) - (1 + x1^2) + 1 )
        #      = x2 sqrt( 1 - (1 + x1^2) (1 - (n2/n1)^2) )
        normsqr = 1 + photon_array.dxdz**2 + photon_array.dydz**2  # (1 + x1^2)
        with np.errstate(invalid='ignore'):
            # NaN below <=> total internal reflection
            factor = np.sqrt(1 - normsqr*(1-index_ratio**2))
        photon_array.dxdz /= factor
        photon_array.dydz /= factor
        photon_array.flux = np.where(np.isnan(factor), 0.0, photon_array.flux)

    def __repr__(self):
        return "galsim.Refraction(index_ratio=%r)"%self.index_ratio


class FocusDepth(PhotonOp):
    """A photon operator that focuses/defocuses photons by changing the height of the focal
    surface with respect to the conical beam.

    Parameters:
        depth:   The z-distance by which to displace the focal surface, in units of pixels.  A
                 positive (negative) number here indicates an extra- (intra-) focal sensor height.
                 I.e., depth > 0 means the sensor surface intersects the beam after it has
                 converged, and depth < 0 means the sensor surface intersects the beam before it
                 has converged.
    """
    _req_params = { 'depth' : float }

    def __init__(self, depth):
        self.depth = depth

    def applyTo(self, photon_array, local_wcs=None, rng=None):
        """Adjust a photon bundle to account for the change in focal depth.

        Parameters:
            photon_array:   A `PhotonArray` to apply the operator to.
            local_wcs:      A `LocalWCS` instance defining the local WCS for the current photon
                            bundle in case the operator needs this information.  [default: None]
            rng:            A random number generator to use if needed. [default: None]
        """
        if not photon_array.hasAllocatedAngles():
            raise GalSimError("FocusDepth requires that angles be set")
        photon_array.x += self.depth * photon_array.dxdz
        photon_array.y += self.depth * photon_array.dydz

    def __repr__(self):
        return "galsim.FocusDepth(depth=%r)"%self.depth


class PupilImageSampler(PhotonOp):
    """A photon operator that samples the pupil-plane positions given a pupil-plane image.
    Samples are drawn discretely from pupil plane image pixels marked as illuminated.

    Parameters:
        diam:               Aperture diameter in meters.
        lam:                Wavelength in nanometers.  [default: None]
        circular_pupil:     Adopt a circular pupil?  [default: True]
        obscuration:        Linear dimension of central obscuration as fraction of aperture
                            linear dimension. [0., 1.).  [default: 0.0]
        nstruts:            Number of radial support struts to add to the central obscuration.
                            [default: 0]
        strut_thick:        Thickness of support struts as a fraction of aperture diameter.
                            [default: 0.05]
        strut_angle:        `Angle` made between the vertical and the strut starting closest to it,
                            defined to be positive in the counter-clockwise direction; must be an
                            `Angle` instance. [default: 0. * galsim.degrees]
        oversampling:       Optional oversampling factor *in the image plane* for the PSF
                            eventually constructed using this `Aperture`.  Setting
                            ``oversampling < 1`` will produce aliasing in the PSF (not good).
                            [default: 1.0]
        pad_factor:         Additional multiple by which to extend the PSF image to avoid
                            folding.  [default: 1.0]
        pupil_plane_im:     The GalSim.Image, NumPy array, or name of file containing the pupil
                            plane image, to be used instead of generating one based on the
                            obscuration and strut parameters.  [default: None]
        pupil_angle:        If ``pupil_plane_im`` is not None, rotation angle for the pupil plane
                            (positive in the counter-clockwise direction).  Must be an `Angle`
                            instance. [default: 0. * galsim.degrees]
        pupil_plane_scale:  Sampling interval in meters to use for the pupil plane array.  In
                            most cases, it's a good idea to leave this as None, in which case
                            GalSim will attempt to find a good value automatically.  The
                            exception is when specifying the pupil arrangement via an image, in
                            which case this keyword can be used to indicate the sampling of that
                            image.  See also ``pad_factor`` for adjusting the pupil sampling scale.
                            [default: None]
        pupil_plane_size:   Size in meters to use for the pupil plane array.  In most cases, it's
                            a good idea to leave this as None, in which case GalSim will attempt
                            to find a good value automatically.  See also ``oversampling`` for
                            adjusting the pupil size.  [default: None]
    """
    _req_params = {
        "diam": float,
    }
    _opt_params = {
        "lam": float,
        "circular_pupil": bool,
        "obscuration": float,
        "nstruts": int,
        "strut_thick": float,
        "strut_angle": Angle,
        "oversampling": float,
        "pad_factor": float,
        "pupil_plane_im": str,
        "pupil_angle": Angle,
        "pupil_plane_scale": float,
        "pupil_plane_size": float,
    }
    def __init__(self, diam, **kwargs):
        from .phase_psf import Aperture
        self.aper = Aperture(diam, **kwargs)
        # Save these for the repr
        self.diam = diam
        self.kwargs = kwargs

    def applyTo(self, photon_array, local_wcs=None, rng=None):
        """Sample the pupil plane u,v positions for each photon.

        Parameters:
            photon_array:   A `PhotonArray` to apply the operator to.
            local_wcs:      A `LocalWCS` instance defining the local WCS for the current photon
                            bundle in case the operator needs this information.  [default: None]
            rng:            A random number generator to use if needed. [default: None]
        """
        self.aper.samplePupil(photon_array, rng)

    def __repr__(self):
        s =  "galsim.PupilImageSampler(diam=%s"%self.diam
        for k,v in self.kwargs.items():
            s += ', %s=%r'%(k,v)
        s += ')'
        return s


class PupilAnnulusSampler(PhotonOp):
    """A photon operator that uniformly samples an annular entrance pupil.

    Parameters:
        R_outer:  Annulus outer radius in meters.
        R_inner:  Annulus inner radius in meters.  [default: 0.0]
    """
    _req_params = {
        "R_outer": float,
    }
    _opt_params = {
        "R_inner": float,
    }
    def __init__(self, R_outer, R_inner=0.0):
        self.R_outer = R_outer
        self.R_inner = R_inner

    def applyTo(self, photon_array, local_wcs=None, rng=None):
        """Sample the pupil plane u,v positions for each photon.

        Parameters:
            photon_array:   A `PhotonArray` to apply the operator to.
            local_wcs:      A `LocalWCS` instance defining the local WCS for the current photon
                            bundle in case the operator needs this information.  [default: None]
            rng:            A random number generator to use if needed. [default: None]
        """
        gen = BaseDeviate(rng).as_numpy_generator()
        r = gen.uniform(self.R_inner**2, self.R_outer**2, size=len(photon_array))
        np.sqrt(r, out=r)
        phi = gen.uniform(0, 2*np.pi, size=len(photon_array))
        photon_array.pupil_u = r * np.cos(phi)
        photon_array.pupil_v = r * np.sin(phi)

    def __repr__(self):
        s = "galsim.PupilAnnulusSampler(R_outer=%r"%self.R_outer
        if self.R_inner != 0.0:
            s += ", R_inner=%r"%self.R_inner
        s += ")"
        return s


class TimeSampler(PhotonOp):
    """A photon operator that uniformly samples photon time stamps within some interval.

    Parameters:
        t0:         The nominal start time of the observation in seconds. [default: 0]
        exptime:    The exposure time in seconds. [default: 0]
    """
    _opt_params = {
        "t0": float,
        "exptime": float
    }
    def __init__(self, t0=0.0, exptime=0.0):
        self.t0 = t0
        self.exptime = exptime

    def applyTo(self, photon_array, local_wcs=None, rng=None):
        """Add time stamps to photons.

        Parameters:
            photon_array:   A `PhotonArray` to apply the operator to.
            local_wcs:      A `LocalWCS` instance defining the local WCS for the current photon
                            bundle in case the operator needs this information.  [default: None]
            rng:            A random number generator to use if needed. [default: None]
        """
        gen = BaseDeviate(rng).as_numpy_generator()
        photon_array.time = gen.uniform(self.t0, self.t0+self.exptime, size=len(photon_array))

    def __repr__(self):
        s = "galsim.TimeSampler("
        if self.t0 != 0.0:
            s += "t0=%r"%self.t0
            if self.exptime != 0.0:
                s += ", exptime=%r"%self.exptime
        else:
            if self.exptime != 0.0:
                s += "exptime=%r"%self.exptime
        s += ")"
        return s


class ScaleFlux(PhotonOp):
    """A simple photon operator that multiplies all flux values by a constant.

    Parameters:
        x:          The constant by which to multiply all flux values.
    """
    _req_params = { "x": float }
    def __init__(self, x):
        self.x = x

    def applyTo(self, photon_array, local_wcs=None, rng=None):
        """Apply the scaling.
        """
        photon_array.flux *= self.x

    def __repr__(self):
        return f"galsim.ScaleFlux({self.x})"


class ScaleWavelength(PhotonOp):
    """A simple photon operator that multiplies all wavelength values by a constant.

    Parameters:
        x:          The constant by which to multiply all wavelength values.
    """
    _req_params = { "x": float }
    def __init__(self, x):
        self.x = x

    def applyTo(self, photon_array, local_wcs=None, rng=None):
        """Apply the scaling.
        """
        photon_array.wavelength *= self.x

    def __repr__(self):
        return f"galsim.ScaleWavelength({self.x})"


# Put these at the end to avoid circular imports
from . import fits



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

import time
import numpy as np

import galsim
from galsim_test_helpers import *

VAR_NDECIMAL=4
CHECKNOISE_NDECIMAL=2

@timer
def test_metacal_tracking():
    """Test that the noise tracking works for the metacal use case involving deconvolution and
    reconvolution by almost the same PSF.

    This test is similar to the above test_uncorrelated_noise_tracking, except the modifications
    are based on what is done for the metacal procedure.
    """
    import math
    def check_noise(noise_image, noise, msg):
        # A helper function to check that the current noise in the image is properly described
        # by the given CorrelatedNoise object
        noise2 = galsim.CorrelatedNoise(noise_image)
        print('noise = ',noise)
        print('noise2 = ',noise2)
        np.testing.assert_almost_equal(noise.getVariance(), noise2.getVariance(),
                                       decimal=CHECKNOISE_NDECIMAL,
                                       err_msg=msg + ': variance does not match.')
        cf_im1 = galsim.Image(8,8, wcs=noise_image.wcs)
        cf_im2 = cf_im1.copy()
        noise.drawImage(image=cf_im1)
        noise2.drawImage(image=cf_im2)
        np.testing.assert_almost_equal(cf_im1.array, cf_im2.array,
                                       decimal=CHECKNOISE_NDECIMAL,
                                       err_msg=msg + ': image of cf does not match.')

    def check_symm_noise(noise_image, msg):
        # A helper funciton to see if a noise image has 4-fold symmetric noise.
        im2 = noise_image.copy()
        # Clear out any wcs to make the test simpler
        im2.wcs = galsim.PixelScale(1.)
        noise = galsim.CorrelatedNoise(im2)
        cf = noise.drawImage(galsim.Image(bounds=galsim.BoundsI(-1,1,-1,1), scale=1))
        # First check the variance
        print('variance: ',cf(0,0), noise.getVariance())
        np.testing.assert_almost_equal(cf(0,0)/noise.getVariance(), 1.0, decimal=VAR_NDECIMAL,
                                       err_msg=msg + ':: noise variance is wrong.')
        cf_plus = np.array([ cf(1,0), cf(-1,0), cf(0,1), cf(0,-1) ])
        cf_cross = np.array([ cf(1,1), cf(-1,-1), cf(-1,1), cf(1,-1) ])
        print('plus pattern: ',cf_plus)
        print('diff relative to dc: ',(cf_plus-np.mean(cf_plus))/cf(0,0))
        print('cross pattern: ',cf_cross)
        print('diff relative to dc: ',(cf_cross-np.mean(cf_cross))/cf(0,0))
        # For now, don't make these asserts.  Just print whether they will pass or fail.
        if True:
            if np.all(np.abs((cf_plus-np.mean(cf_plus))/cf(0,0)) < 0.01):
                print('plus test passes')
            else:
                print('*** FAIL ***')
                print(msg + ': plus pattern is not constant')
            if np.all(np.abs((cf_cross-np.mean(cf_cross))/cf(0,0)) < 0.01):
                print('cross test passes')
            else:
                print('*** FAIL ***')
                print(msg + ': cross pattern is not constant')
        else:
            np.testing.assert_almost_equal((cf_plus-np.mean(cf_plus))/cf(0,0), 0.0, decimal=2,
                                           err_msg=msg + ': plus pattern is not constant')
            np.testing.assert_almost_equal((cf_cross-np.mean(cf_cross))/cf(0,0), 0.0, decimal=2,
                                           err_msg=msg + ': cross pattern is not constant')

    seed = 1234567  # For use as a unit test, we need a specific seed
    #seed = 0  # During testing, it's useful to see how numbers flop around to know if
               # something is systematic or random.
    rng = galsim.BaseDeviate(seed)

    noise_var = 1.3
    im_size = 1024
    dg = 0.1  # This is bigger than metacal would use, but it makes the test easier.

    # Use a non-trivial wcs...
    #wcs = galsim.JacobianWCS(0.26, 0.04, -0.04, -0.26)  # Rotation + flip.  No shear.
    wcs = galsim.JacobianWCS(0.26, 0.03, 0.08, -0.21)   # Fully complex
    #dudx =  0.12*0.26
    #dudy =  1.10*0.26
    #dvdx = -0.915*0.26
    #dvdy = -0.04*0.26
    #wcs = galsim.JacobianWCS(dudx, dudy, dvdx, dvdy)  # Even more extreme

    # And an asymmetric PSF
    #orig_psf = galsim.Gaussian(fwhm=0.9).shear(g1=0.05, g2=0.03)
    # This one is small enough not to be fully Nyquist sampled, which makes things harder.
    orig_psf = galsim.Gaussian(fwhm=0.7).shear(g1=0.05, g2=0.03)

    # pixel is the pixel in world coords
    pixel = wcs.toWorld(galsim.Pixel(scale=1))
    pixel_inv = galsim.Deconvolve(pixel)

    psf_image = orig_psf.drawImage(nx=im_size, ny=im_size, wcs=wcs)

    # Metacal only has access to the PSF as an image, so use this from here on.
    psf = galsim.InterpolatedImage(psf_image)
    psf_nopix = galsim.Convolve([psf, pixel_inv])
    psf_inv = galsim.Deconvolve(psf)

    # Not what is done currently, but using a smoother target PSF helps make sure the
    # zeros in the deconvolved PSF get adequately zeroed out.
    def get_target_psf(psf):
        dk = 0.1              # The resolution in k space for the KImage
        small_kval = 1.e-2    # Find the k where the given psf hits this kvalue
        smaller_kval = 3.e-3  # Target PSF will have this kvalue at the same k

        kim = psf.drawKImage(scale=dk)
        karr_r = kim.real.array
        # Find the smallest r where the kval < small_kval
        nk = karr_r.shape[0]
        kx, ky = np.meshgrid(np.arange(-nk/2,nk/2), np.arange(-nk/2,nk/2))
        ksq = (kx**2 + ky**2) * dk**2
        ksq_max = np.min(ksq[karr_r < small_kval * psf.flux])

        # We take our target PSF to be the (round) Gaussian that is even smaller at this ksq
        # exp(-0.5 * ksq_max * sigma_sq) = smaller_kval
        sigma_sq = -2. * np.log(smaller_kval) / ksq_max
        return galsim.Gaussian(sigma = np.sqrt(sigma_sq))

    # The target PSF dilates the part without the pixel, but reconvolve by the real pixel.
    #psf_target_nopix = psf_nopix.dilate(1. + 2.*dg)
    #psf_target_nopix = orig_psf.dilate(1. + 4.*dg)
    psf_target_nopix = get_target_psf(psf_nopix.shear(g1=dg))
    print('PSF target HLR = ',psf_target_nopix.calculateHLR())
    print('PSF target FWHM = ',psf_target_nopix.calculateFWHM())
    psf_target = galsim.Convolve([psf_target_nopix, pixel])

    # Make an image of pure (white) Gaussian noise
    # Normally, there would be a galaxy in this image, but for the tests, we just have noise.
    obs_image = galsim.Image(im_size, im_size, init_value=0, wcs=wcs)
    obs_image.addNoise(galsim.GaussianNoise(rng=rng, sigma=math.sqrt(noise_var)))

    # The noise on this image should be describable as an UncorrelatedNoise object:
    noise = galsim.UncorrelatedNoise(variance=noise_var, wcs=wcs)
    check_noise(obs_image, noise, 'initial UncorrelatedNoise model is wrong')

    # Make an InterpolatedImage profile to use for manipulating this image
    # We can get away with no padding here, since our image is so large, but normally, you would
    # probably want to pad this with noise padding.
    ii = galsim.InterpolatedImage(obs_image, pad_factor=1)
    ii.noise = noise

    # If we draw it back, the attached noise attribute should still be correct
    check_noise(ii.drawImage(obs_image.copy(),method='no_pixel'), ii.noise,
                'noise model is wrong for InterpolatedImage')

    # Here is the metacal process for which we want to understand the noise.
    # We'll try a few different methods.
    shear = galsim.Shear(g1=dg)
    sheared_obj = galsim.Convolve(ii, psf_inv).shear(shear)
    final_obj = galsim.Convolve(psf_target, sheared_obj)
    final_image = final_obj.drawImage(obs_image.copy(), method='no_pixel')

    try:
        check_symm_noise(final_image, 'Initial image')
        #didnt_fail = True
        # This bit doesn't work while we are not actually raising exceptions in check_symm_noise
        # So we expect to see **FAIL** text at this point.
        print('The above tests are expected to **FAIL**.  This is not a problem.')
        didnt_fail = False
    except AssertionError as e:
        print('As expected initial image fails symmetric noise test:')
        print(e)
        didnt_fail = False
    if didnt_fail:
        assert False, 'Initial image was expected to fail symmetric noise test, but passed.'

    if True:
        print('\n\nStrategy 1:')
        # Strategy 1: Use the noise attribute attached to ii and use it to either whiten or
        #             symmetrize the noise in the final image.
        # Note: The check_noise tests fail.  I think because the convolve and deconvolve impose
        #       a maxk = that of the psf.  Which is too small for an accurate rendering of the
        #       correlation function (here just an autocorrelation of a Pixel.
        # The whiten tests kind of work, but they add a lot of extra noise.  Much more than
        # strategy 4 below.  So the level of correlation remaining is pretty well below the
        # dc variance.  Symmetrize doesn't add much noise, but the residual correlation is about
        # the same, which means it doesn't pass the test relative to the lower dc variance.

        # First, deconvolve and reconvolve by the same PSF:
        test_obj = galsim.Convolve([ii, psf, psf_inv])
        # This fails...
        if False:
            check_noise(test_obj.drawImage(obs_image.copy(), method='no_pixel'), test_obj.noise,
                        'noise model is wrong after convolve/deconvolve by psf')

        # Now use a slightly dilated PSF for the reconvolution:
        test_obj = galsim.Convolve([ii, psf_target, psf_inv])
        if False:
            check_noise(test_obj.drawImage(obs_image.copy(), method='no_pixel'), test_obj.noise,
                        'noise model is wrong for dilated target psf')

        # Finally, include the shear step.  This was done above with sheared_obj, final_obj.
        if False:
            check_noise(final_image, final_obj.noise,
                        'noise model is wrong when including small shear')

        # If we whiten using this noise model, we should get back to white noise.
        t3 = time.time()
        final_image2 = final_image.copy()  # Don't clobber the original
        final_var = final_image2.whitenNoise(final_obj.noise)
        t4 = time.time()
        print('Noise tracking method with whiten: final_var = ',final_var)
        print('Check: direct variance = ',np.var(final_image2.array))
        check_symm_noise(final_image2, 'noise whitening does not work')
        print('Time for noise tracking with whiten = ',t4-t3)

        # Using symmetrizeNoise should add less noise, but also work.
        t3 = time.time()
        final_image2 = final_image.copy()
        final_var = final_image2.symmetrizeNoise(final_obj.noise)
        t4 = time.time()
        print('Noise tracking method with symmetrize: final_var = ',final_var)
        print('Check: direct variance = ',np.var(final_image2.array))
        check_symm_noise(final_image2, 'noise symmetrizing does not work')
        print('Time for noise tracking with symmetrize = ',t4-t3)


    if True:
        print('\n\nStrategy 2:')
        # Strategy 2: Don't trust the noise tracking. Track a noise image through the same process
        #             and then measure the noise from that image.  Use it to either whiten or
        #             symmetrize the noise in the final image.
        # Note: This method doesn't work any better.  The added noise for whitening is even more
        # than strategy 1.  And the remaining correlations are still similarly significant for the
        # symmetrize version.  A little smaller than strategy 1, but not enough to pass our tests.

        # Make another noise image, since we don't actually have access to a pure noise image
        # for real objects.  But we should be able to estimate the variance in the image.
        t3 = time.time()
        noise_image = galsim.Image(im_size,im_size, init_value=0, wcs=wcs)
        noise_image.addNoise(galsim.GaussianNoise(rng=rng, sigma=math.sqrt(noise_var)))
        noise_ii = galsim.InterpolatedImage(noise_image, pad_factor=1)
        sheared_noise_obj = galsim.Convolve(noise_ii, psf_inv).shear(shear)
        final_noise_obj = galsim.Convolve(psf_target, sheared_noise_obj)
        final_noise_image = final_noise_obj.drawImage(obs_image.copy(), method='no_pixel')

        # Use this to construct an appropriate CorrelatedNoise object
        noise = galsim.CorrelatedNoise(final_noise_image)
        t4 = time.time()
        final_image2 = final_image.copy()
        final_var = final_image2.whitenNoise(noise)
        t5 = time.time()

        check_noise(final_noise_image, noise,
                    'noise model is wrong when direct measuring the final noise image')

        print('Direct noise method with whiten: final_var = ',final_var)
        # Neither of these work currently, so maybe a bug in the whitening code?
        # Or possibly in my logic here.
        check_symm_noise(final_image2, 'whitening the noise using direct noise model failed')
        print('Time for direct noise with whitening = ',t5-t3)

        t6 = time.time()
        final_image2 = final_image.copy()
        final_var = final_image2.symmetrizeNoise(noise)
        t7 = time.time()

        print('Direct noise method with symmetrize: final_var = ',final_var)
        check_symm_noise(final_image2, 'symmetrizing the noise using direct noise model failed')
        print('Time for direct noise with symmetrizing = ',t7-t6 + t4-t3)

    if False:
        print('\n\nStrategy 3:')
        # Strategy 3: Make a noise field and do the same operations as we do to the main image
        #             except use the opposite shear value.  Add this noise field to the final
        #             image to get a symmetric noise field.
        # Note: This method works!  But only for square pixels.  However, they may be rotated
        # or flipped. Just not sheared.
        # Update: I think this method won't ever work for non-square pixels.  The reason it works
        # for square pixels is that in that case, it is equivalent to strategy 4.
        t3 = time.time()

        # Make another noise image
        rev_image = galsim.Image(im_size,im_size, init_value=0, wcs=wcs)
        rev_image.addNoise(galsim.GaussianNoise(rng=rng, sigma=math.sqrt(noise_var)))
        rev_ii = galsim.InterpolatedImage(rev_image, pad_factor=1)

        rev_sheared_obj = galsim.Convolve(rev_ii, psf_inv).shear(-shear)
        rev_final_obj = galsim.Convolve(psf_target, rev_sheared_obj)
        rev_final_image = rev_final_obj.drawImage(obs_image.copy(), method='no_pixel')

        # Add the reverse-sheared noise image to the original image.
        final_image2 = final_image + rev_final_image
        t4 = time.time()

        # The noise variance in the end should be 2x as large as the original
        final_var = np.var(final_image2.array)
        print('Reverse shear method: final_var = ',final_var)
        check_symm_noise(final_image2, 'using reverse shear does not work')
        print('Time for reverse shear method = ',t4-t3)

    if True:
        print('\n\nStrategy 4:')
        # Strategy 4: Make a noise field and do the same operations as we do to the main image,
        #             then rotate it by 90 degress and add it to the final image.
        # This method works!  Even for an arbitrarily sheared wcs.
        t3 = time.time()

        # Make another noise image
        noise_image = galsim.Image(im_size,im_size, init_value=0, wcs=wcs)
        noise_image.addNoise(galsim.GaussianNoise(rng=rng, sigma=math.sqrt(noise_var)))
        noise_ii = galsim.InterpolatedImage(noise_image, pad_factor=1)

        noise_sheared_obj = galsim.Convolve(noise_ii, psf_inv).shear(shear)
        noise_final_obj = galsim.Convolve(psf_target, noise_sheared_obj)
        noise_final_image = noise_final_obj.drawImage(obs_image.copy(), method='no_pixel')

        # Rotate the image by 90 degrees
        rot_noise_final_image = galsim.Image(np.ascontiguousarray(np.rot90(noise_final_image.array)))

        # Add the rotated noise image to the original image.
        final_image2 = final_image + rot_noise_final_image
        t4 = time.time()

        # The noise variance in the end should be 2x as large as the original
        final_var = np.var(final_image2.array)
        print('Rotate image method: final_var = ',final_var)
        check_symm_noise(final_image2, 'using rotated noise image does not work')
        print('Time for rotate image method = ',t4-t3)

    if False:
        print('\n\nStrategy 5:')
        # Strategy 5: The same as strategy 3, except we target the effective net transformation
        #             done by strategy 4.
        # I think this strategy probably can't work for non-square pixels, because the shear
        # happens before the convolution by the PSF.  And if the wcs is non-square, then the
        # PSF is sheared relative to the pixels.  That shear isn't being accounted for here,
        # so the net result isn't equivalent to rotating by 90 degrees at the end.
        t3 = time.time()

        # Make another noise image
        rev_image = galsim.Image(im_size,im_size, init_value=0, wcs=wcs)
        rev_image.addNoise(galsim.GaussianNoise(rng=rng, sigma=math.sqrt(noise_var)))
        rev_ii = galsim.InterpolatedImage(rev_image, pad_factor=1)

        # Find the effective transformation to apply in sky coordinates that matches what
        # you would get by applying the shear in sky coords, going to image coords and then
        # rotating by 90 degrees.
        #
        # If J is the jacobian of the wcs, and S1 is the applied shear, then we want to find
        # S2 such that J^-1 S2 = R90 J^-1 S1
        jac = wcs.jacobian()
        J = jac.getMatrix()
        Jinv = jac.inverse().getMatrix()
        S1 = shear.getMatrix()
        R90 = np.array([[0,-1],[1,0]])
        S2 = J.dot(R90).dot(Jinv).dot(S1)
        scale, rev_shear, rev_theta, flip = galsim.JacobianWCS(*S2.flatten()).getDecomposition()
        # Flip should be False, and scale should be essentially 1.0.
        assert flip == False
        assert abs(scale - 1.) < 1.e-8

        rev_sheared_obj = galsim.Convolve(rev_ii, psf_inv).rotate(rev_theta).shear(rev_shear)
        rev_final_obj = galsim.Convolve(psf_target, rev_sheared_obj)
        rev_final_image = rev_final_obj.drawImage(obs_image.copy(), method='no_pixel')

        # Add the reverse-sheared noise image to the original image.
        final_image2 = final_image + rev_final_image
        t4 = time.time()

        # The noise variance in the end should be 2x as large as the original
        final_var = np.var(final_image2.array)
        print('Alternate reverse shear method: final_var = ',final_var)
        check_symm_noise(final_image2, 'using alternate reverse shear does not work')
        print('Time for alternate reverse shear method = ',t4-t3)

@timer
def test_wcs():
    """Reproduce an error Erin found and reported in #834.  This was never an error in a released
    version, just temporarily on master, but the mistake hadn't been caught by any of our unit
    tests, so this test catches the error.
    """
    wcs = galsim.JacobianWCS(0.01, -0.26, -0.28, -0.03)
    gal = galsim.Exponential(half_light_radius=1.1, flux=237)
    psf = galsim.Moffat(beta=3.5, half_light_radius=0.9)
    obs = galsim.Convolve(gal,psf)

    obs_im = obs.drawImage(nx=32, ny=32, offset=(0.3,-0.2), wcs=wcs)
    psf_im = psf.drawImage(nx=32, ny=32, wcs=wcs)

    ii = galsim.InterpolatedImage(obs_im)
    psf_ii = galsim.InterpolatedImage(psf_im)
    psf_inv = galsim.Deconvolve(psf_ii)

    ii_nopsf = galsim.Convolve(ii, psf_inv)

    newpsf=galsim.Moffat(beta=3.5, half_light_radius=0.95)
    newpsf=newpsf.dilate(1.02)

    new_ii = ii_nopsf.shear(g1=0.01,g2=0.0)
    new_ii = galsim.Convolve(new_ii, newpsf)

    new_im = new_ii.drawImage(image=obs_im.copy(), method='no_pixel')
    np.testing.assert_almost_equal(new_im.array.sum()/237, obs_im.array.sum()/237, decimal=1)

if __name__ == "__main__":
    runtests(__file__)



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

import numpy as np
import warnings

import galsim
from galsim_test_helpers import *

with warnings.catch_warnings():
    warnings.filterwarnings("ignore",category=RuntimeWarning)
    import scipy.special

@timer
def test_j0():
    """Test the bessel.j0 function"""
    x_list = [ 0, 1.01, 0.2, 3.3, 5.9, 77., 1.e-12, 500. ]
    vals1 = [ galsim.bessel.j0(x) for x in x_list ]
    print('x = ',x_list)
    print('vals1 = ',vals1)

    vals2 = [ scipy.special.j0(x) for x in x_list ]
    print('vals2 = ',vals2)
    np.testing.assert_allclose(
        vals1, vals2, rtol=1.e-10, err_msg="bessel.j0 disagrees with scipy.special.j0")

    # Scipy is now considered required for tests, so these explicit values are not necessary
    # anymore as a backup, but I'm leaving them in as a regression test.  They should be the same
    # values that scipy returns above.
    vals2 = [   1.0,
                0.76078097763218844,
                0.99002497223957631,
                -0.34429626039888467,
                0.12203335459282282,
                0.062379777089647245,
                1.0,
                -0.034100556880731728,
            ]
    np.testing.assert_allclose(
        vals1, vals2, rtol=1.e-10, err_msg="bessel.j0 disagrees with reference values")


@timer
def test_j1():
    """Test the bessel.j1 function"""
    x_list = [ 0, 1.01, 0.2, 3.3, 5.9, 77., 1.e-12, 500. ]
    vals1 = [ galsim.bessel.j1(x) for x in x_list ]
    print('x = ',x_list)
    print('vals1 = ',vals1)

    vals2 = [ scipy.special.j1(x) for x in x_list ]
    print('vals2 = ',vals2)
    np.testing.assert_allclose(
        vals1, vals2, rtol=1.e-10, err_msg="bessel.j1 disagrees with scipy.special.j1")

    # These values are what scipy returns.
    vals2 = [   0.0,
                0.4432857612090717,
                0.099500832639236036,
                0.22066345298524112,
                -0.29514244472901613,
                0.066560642470571682,
                5.0000000000000009e-13,
                0.010472613470372989,
            ]
    np.testing.assert_allclose(
        vals1, vals2, rtol=1.e-10, err_msg="bessel.j1 disagrees with reference values")


@timer
def test_jn():
    """Test the bessel.jn function"""
    n_list = [ 3, 4, 1, 0, 9, -7, 4, 300, 39, 2 ]
    x_list = [ 0, 1.01, 0.2, 3.3, 15.9, 77., 1.e-12, 500., 3.4, 19.1 ]
    vals1 = [ galsim.bessel.jn(n,x) for n,x in zip(n_list,x_list) ]
    print('x = ',x_list)
    print('vals1 = ',vals1)

    vals2 = [ scipy.special.jn(n,x) for n,x in zip(n_list,x_list) ]
    print('vals2 = ',vals2)
    np.testing.assert_allclose(
        vals1, vals2, rtol=1.e-10, err_msg="bessel.jn disagrees with scipy.special.jn")

    # These values are what scipy returns.
    vals2 = [   0.0,
                0.0025745895535573995,
                0.099500832639236036,
                -0.34429626039888467,
                -0.19886654556845093,
                0.082526868218916541,
                2.6041666666666486e-51,
                -0.0029540008506870503,
                4.4311593707049111e-38,
                -0.16584939097852849,
            ]
    np.testing.assert_allclose(
        vals1, vals2, rtol=1.e-10, err_msg="bessel.jn disagrees with reference values")


@timer
def test_jv():
    """Test the bessel.jv function"""
    v_list = [ 3.3, 4, 1.9, 0, 0.2, -7.1, 4.7, 300.9, 39.8, 2.3 ]
    x_list = [ 0, 1.01, 0.2, 3.3, 15.9, 77., 1.e-12, 500., 3.4, 19.1 ]
    vals1 = [ galsim.bessel.jv(v,x) for v,x in zip(v_list,x_list) ]
    print('x = ',x_list)
    print('vals1 = ',vals1)

    vals2 = [ scipy.special.jv(v,x) for v,x in zip(v_list,x_list) ]
    print('vals2 = ',vals2)
    np.testing.assert_allclose(
        vals1, vals2, rtol=1.e-10, err_msg="bessel.jv disagrees with scipy.special.jv")

    # These values are what scipy returns.
    vals2 = [   0.0,
                0.0025745895535573995,
                0.0068656051839294848,
                -0.34429626039888467,
                -0.12213323547609471,
                0.087784805831697565,
                2.1118132341112502e-60,
                0.027491061612400482,
                3.5539536207995954e-39,
                -0.11753152687355309,
            ]
    np.testing.assert_allclose(
        vals1, vals2, rtol=1.e-10, err_msg="bessel.jv disagrees with reference values")


@timer
def test_yn():
    """Test the bessel.yn function"""
    n_list = [ 3, 4, 1, 0, 9, -7, 4, 300, 39, 2 ]
    x_list = [ 0.1, 1.01, 0.2, 3.3, 15.9, 77., 1.e-12, 500., 3.4, 19.1 ]
    vals1 = [ galsim.bessel.yn(n,x) for n,x in zip(n_list,x_list) ]
    print('x = ',x_list)
    print('vals1 = ',vals1)

    vals2 = [ scipy.special.yn(n,x) for n,x in zip(n_list,x_list) ]
    print('vals2 = ',vals2)
    np.testing.assert_allclose(
        vals1, vals2, rtol=1.e-10, err_msg="bessel.yn disagrees with scipy.special.yn")

    # These values are what scipy returns.
    vals2 = [   -5099.3323786129049,
                -32.036214020138011,
                -3.3238249881118471,
                0.26909199505453402,
                -0.094117150977372527,
                -0.038617322619578592,
                -3.0557749073643908e+49,
                0.039784618934117735,
                -1.8489533038779474e+35,
                0.077438881119458292,
            ]
    np.testing.assert_allclose(
        vals1, vals2, rtol=1.e-10, err_msg="bessel.yn disagrees with reference values")


@timer
def test_yv():
    """Test the bessel.yv function"""
    v_list = [ 3.3, 4, 1.9, 0, 0.2, -7.1, 4.7, 300.9, 39.8, 2.3 ]
    x_list = [ 0.1, 1.01, 0.2, 3.3, 15.9, 77., 1.e-12, 500., 3.4, 19.1 ]
    vals1 = [ galsim.bessel.yv(v,x) for v,x in zip(v_list,x_list) ]
    print('x = ',x_list)
    print('vals1 = ',vals1)

    vals2 = [ scipy.special.yv(v,x) for v,x in zip(v_list,x_list) ]
    print('vals2 = ',vals2)
    np.testing.assert_allclose(
        vals1, vals2, rtol=1.e-10, err_msg="bessel.yv disagrees with scipy.special.yv")

    # These values are what scipy returns.
    vals2 = [   -16804.006307563286,
                -32.036214020138011,
                -24.595386714889109,
                0.26909199505453379,
                0.15844850776917807,
                -0.024429586880444932,
                -3.2069837713267924e+58,
                0.028956519948967349,
                -2.2586390441339445e+36,
                0.14053075843601004,
            ]
    np.testing.assert_allclose(
        vals1, vals2, rtol=1.e-10, err_msg="bessel.yv disagrees with reference values")

@timer
def test_in():
    """Test the bessel.iv function with integer values.  (There is no bessel.in, nor
    scipy.special.in, since in is of course a reserved word.)
    """
    n_list = [ 3, 4, 1, 0, 9, 7, 4, 300, 39, 2 ]
    x_list = [ 0, 2.01, 0.2, 3.3, 15.9, 7.7, 1.e-12, 500., 3.4, 19.1 ]
    vals1 = [ galsim.bessel.iv(n,x) for n,x in zip(n_list,x_list) ]
    print('x = ',x_list)
    print('vals1 = ',vals1)

    vals2 = [ scipy.special.iv(n,x) for n,x in zip(n_list,x_list) ]
    print('vals2 = ',vals2)
    np.testing.assert_allclose(
        vals1, vals2, rtol=1.e-10, err_msg="bessel.iv disagrees with scipy.special.iv")

    # These values are what scipy returns.
    vals2 = [   0.0,
                0.051851345436838572,
                0.10050083402812511,
                6.2426304651830256,
                62790.244278205071,
                13.590333620044147,
                2.6041666666666664e-51,
                2.1971573261481075e+177,
                5.1200364123222137e-38,
                16279753.373047709,
            ]
    np.testing.assert_allclose(
        vals1, vals2, rtol=1.e-10, err_msg="bessel.iv disagrees with reference values")



@timer
def test_iv():
    """Test the bessel.iv function"""
    v_list = [ 3.3, 4, 1.9, 0, 0.2, -7.1, 4.7, 300.9, 39.8, 2.3 ]
    x_list = [ 0, 2.01, 0.2, 3.3, 15.9, 7.7, 1.e-12, 500., 3.4, 19.1 ]
    vals1 = [ galsim.bessel.iv(v,x) for v,x in zip(v_list,x_list) ]
    print('x = ',x_list)
    print('vals1 = ',vals1)

    vals2 = [ scipy.special.iv(v,x) for v,x in zip(v_list,x_list) ]
    print('vals2 = ',vals2)
    np.testing.assert_allclose(
        vals1, vals2, rtol=1.e-10, err_msg="bessel.iv disagrees with scipy.special.iv")

    # These values are what scipy returns.
    vals2 = [   0.0,
                0.051851345436838572,
                0.0069131178533799404,
                6.2426304651830256,
                809950.48268613976,
                12.476741646263305,
                2.1118132341112364e-60,
                1.315385468943472e+177,
                4.0948398641503255e-39,
                15725726.116463179,
            ]
    np.testing.assert_allclose(
        vals1, vals2, rtol=1.e-10, err_msg="bessel.iv disagrees with reference values")


@timer
def test_kn():
    """Test the bessel.kn function"""
    n_list = [ 3, 4, 1, 0, 9, 7, 4, 300, 39, 2 ]
    x_list = [ 1, 2.01, 0.2, 3.3, 15.9, 7.7, 1.e-12, 500., 3.4, 19.1 ]
    vals1 = [ galsim.bessel.kn(n,x) for n,x in zip(n_list,x_list) ]
    print('x = ',x_list)
    print('vals1 = ',vals1)

    vals2 = [ scipy.special.kn(n,x) for n,x in zip(n_list,x_list) ]
    print('vals2 = ',vals2)
    np.testing.assert_allclose(
        vals1, vals2, rtol=1.e-10, err_msg="bessel.kn disagrees with scipy.special.kn")

    # These values are what scipy returns.
    vals2 = [   7.1012628247379448,
                2.1461917781688697,
                4.7759725432204725,
                0.024610632145839577,
                4.3581406072741754e-07,
                0.0035326394473326195,
                4.8e+49,
                3.902737598763313e-181,
                2.494520930845136e+35,
                1.5997755223851229e-09,
            ]
    np.testing.assert_allclose(
        vals1, vals2, rtol=1.e-10, err_msg="bessel.kn disagrees with reference values")


@timer
def test_kv():
    """Test the bessel.kv function"""
    v_list = [ 3.3, 4, 1.9, 0, 0.2, -7.1, 4.7, 300.9, 39.8, 2.3 ]
    x_list = [ 1, 2.01, 0.2, 3.3, 15.9, 7.7, 1.e-12, 500., 3.4, 19.1 ]
    vals1 = [ galsim.bessel.kv(v,x) for v,x in zip(v_list,x_list) ]
    print('x = ',x_list)
    print('vals1 = ',vals1)

    vals2 = [ scipy.special.kv(v,x) for v,x in zip(v_list,x_list) ]
    print('vals2 = ',vals2)
    np.testing.assert_allclose(
        vals1, vals2, rtol=1.e-10, err_msg="bessel.kv disagrees with scipy.special.kv")

    # These values are what scipy returns.
    vals2 = [   11.898213399340918,
                2.1461917781688693,
                37.787322153212926,
                0.024610632145839324,
                3.8841496808266363e-08,
                0.0038228967734928758,
                5.0375183280909704e+58,
                6.513768858596484e-181,
                3.0568215777387072e+36,
                1.6532345134742898e-09,
            ]
    np.testing.assert_allclose(
        vals1, vals2, rtol=1.e-10, err_msg="bessel.kv disagrees with reference values")


@timer
def test_jv_root():
    """Test the bessel.j0_root and jv_root functions"""
    # Our version uses tabulated values up to 40, so a useful test of the extrapolation
    # requires this to have more than 40 items.
    vals1 = [ galsim.bessel.j0_root(s) for s in range(1,51) ]
    print('vals1 = ',vals1)

    vals2 = scipy.special.jn_zeros(0,50)
    print('vals2 = ',vals2)
    np.testing.assert_allclose(
        vals1, vals2, rtol=1.e-10, err_msg="bessel.j0_root disagrees with scipy.special.jn_zeros")

    vals3 = [ galsim.bessel.jv_root(0,s) for s in range(1,51) ]
    np.testing.assert_allclose(
        vals3, vals2, rtol=1.e-10,
        err_msg="bessel.jv_root disagrees with scipy.special.jn_zeros with n=0")

    for v in vals1:
        np.testing.assert_allclose(galsim.bessel.j0(v), 0., atol=1.e-12)
        np.testing.assert_allclose(scipy.special.j0(v), 0., atol=1.e-12)

    for nu in [0, 1, 7, 13]:
        print('nu = ',nu)
        vals1 = [ galsim.bessel.jv_root(nu,s) for s in range(1,51) ]
        print('vals1 = ',vals1)
        vals2 = scipy.special.jn_zeros(nu,50)
        print('vals2 = ',vals2)
        np.testing.assert_allclose(
            vals1, vals2, rtol=1.e-10,
            err_msg="bessel.jv_root disagrees with scipy.special.jn_zeros with n=%f"%nu)
        for v in vals1:
            np.testing.assert_allclose(galsim.bessel.jv(nu,v), 0., atol=1.e-12)
            np.testing.assert_allclose(scipy.special.jv(nu,v), 0., atol=1.e-12)

    # We can also compute zeros of non-integral order, although scipy cannot.
    # So we can't compare with scipy's answers here.  Just check that they really are zeros.
    for nu in [0.5, 0.001, 1.39, 7.3, 13.9, 0.183]:
        print('nu = ',nu)
        vals1 = [ galsim.bessel.jv_root(nu,s) for s in range(1,51) ]
        print('vals1 = ',vals1)
        for v in vals1:
            np.testing.assert_allclose(galsim.bessel.jv(nu,v), 0., atol=1.e-12)
            np.testing.assert_allclose(scipy.special.jv(nu,v), 0., atol=1.e-12)

    with assert_raises(RuntimeError):
        galsim.bessel.jv_root(-0.3,1)
    with assert_raises(RuntimeError):
        galsim.bessel.jv_root(0.3,-1)

@timer
def test_gammainc():
    """Test the bessel.gammainc function"""
    a_list = [ 3.3, 4, 1.9, 0, 0.2, 4.7, 39.8, 2.3 ]
    x_list = [ 1, 2.01, 0.2, 3.3, 15.9, 1.e-12, 3.4, 19.1 ]
    vals1 = [ galsim.bessel.gammainc(a,x) for a,x in zip(a_list,x_list) ]
    print('x = ',x_list)
    print('vals1 = ',vals1)

    vals2 = [ scipy.special.gammainc(a,x) for a,x in zip(a_list,x_list) ]
    print('vals2 = ',vals2)
    np.testing.assert_allclose(
        vals1, vals2, rtol=1.e-10, err_msg="bessel.gammainc disagrees with scipy.special.gammainc")

    # These values are what scipy returns.
    vals2 = [   0.05336162361853338,
                0.14468550606481548,
                0.022580514695946612,
                1.0,
                0.9999999971717825,
                5.489041152199198e-59,
                1.3292187146528721e-28,
                0.9999997850330204
            ]
    np.testing.assert_allclose(
        vals1, vals2, rtol=1.e-10, err_msg="bessel.gammainc disagrees with reference values")

@timer
def test_sinc():
    """Test the bessel.sinc function"""
    x_list = [ 0, 1, -1, 2.01, 0.2, 3.3, 15.9, -7.7, 1.e-12, 500., 3.4, 19.1 ]
    vals1 = [ galsim.bessel.sinc(x) for x in x_list ]
    print('x = ',x_list)
    print('vals1 = ',vals1)

    vals2 = [ scipy.special.sinc(x) for x in x_list ]
    print('vals2 = ',vals2)
    np.testing.assert_allclose(
        vals1, vals2, rtol=1.e-10, err_msg="bessel.sinc disagrees with scipy.special.sinc")

    # These values are what scipy returns.
    vals2 = [   1.0,
                3.8981718325193755e-17,
                3.8981718325193755e-17,
                0.004974306043335856,
                0.935489283788639,
                -0.07803579012128542,
                -0.006186362535116209,
                -0.03344391005197945,
                1.0,
                -1.0231009598277845e-16,
                -0.0890384386636067,
                -0.005149903890489398
            ]
    np.testing.assert_allclose(
        vals1, vals2, rtol=1.e-10, err_msg="bessel.sinc disagrees with reference values")

@timer
def test_si():
    """Test the bessel.si function"""
    x_list = [ 0, 1, -1, 2, 3.99, 4.0, 4.01, 0.2, 3.3, 15.9, 18, 19, 20, -7.7, 1.e-12, 500. ]
    vals1 = [ galsim.bessel.si(x) for x in x_list ]
    print('x = ',x_list)
    print('vals1 = ',vals1)

    vals2 = [ scipy.special.sici(x)[0] for x in x_list ]
    print('vals2 = ',vals2)
    np.testing.assert_allclose(
        vals1, vals2, rtol=1.e-10, err_msg="bessel.si disagrees with scipy.special.sici")

    # These values are what scipy returns.
    vals2 = [   0.0,
                0.9460830703671831,
                -0.9460830703671831,
                1.605412976802695,
                1.7600892984314866,
                1.758203138949053,
                1.7563053683733343,
                0.19955608852623385,
                1.8480807827952113,
                1.6328040281824159,
                1.5366080968611855,
                1.518630031769364,
                1.5482417010434397,
                -1.5361092381286596,
                1e-12,
                1.5725658822431687
            ]
    np.testing.assert_allclose(
        vals1, vals2, rtol=1.e-10, err_msg="bessel.si disagrees with reference values")


@timer
def test_ci():
    """Test the bessel.ci function"""
    x_list = [ 0, 1, -1, 2, 3.99, 4.0, 4.01, 0.2, 3.3, 15.9, 18, 19, 20, -7.7, 1.e-12, 500. ]
    vals1 = [ galsim.bessel.ci(x) for x in x_list ]
    print('x = ',x_list)
    print('vals1 = ',vals1)

    vals2 = [ scipy.special.sici(x)[1] for x in x_list ]
    print('vals2 = ',vals2)
    np.testing.assert_allclose(
        vals1, vals2, rtol=1.e-10, err_msg="bessel.ci disagrees with scipy.special.sici")

    # These values are what scipy returns.
    vals2 = [   -np.inf,
                0.33740392290096816,
                0.33740392290096816,
                0.422980828774865,
                -0.13933609432530814,
                -0.1409816978869305,
                -0.14260429630144628,
                -1.0422055956727818,
                0.02467828460795851,
                -0.008115782282273786,
                -0.04347510299950101,
                0.0051503710084261295,
                0.044419820845353314,
                0.1222458319118463,
                -27.053805451027014,
                -0.0009320008144042904
            ]
    np.testing.assert_allclose(
        vals1, vals2, rtol=1.e-10, err_msg="bessel.ci disagrees with reference values")



if __name__ == "__main__":
    runtests(__file__)



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

import numpy as np
import os

from ..errors import GalSimBoundsError, GalSimConfigError
from ..position import PositionD
from ..bounds import BoundsD
from ..shapelet import Shapelet
from ..gsparams import GSParams
from ..config import InputLoader, RegisterInputType, RegisterObjectType
from ..config import GetAllParams, GetInputObj, SkipThisObject
from .._pyfits import pyfits

class DES_Shapelet:
    """Class that handles DES files describing interpolated polar shapelet decompositions.
    These are stored as ``*_fitpsf.fits`` files.  They are not used in DES anymore, so this
    class is at best of historical interest

    The shapelet PSFs measure a shapelet decomposition of each star and interpolate the shapelet
    coefficients over the image positions.

    Unlike PSFEx, these PSF models are built directly in world coordinates.  The shapelets know
    about the WCS, so they are able to fit the shapelet model directly in terms of arcsec.
    Thus, the getPSF function always returns a profile in world coordinates.

    Typical usage:

        >>> des_shapelet = galsim.des.DES_Shapelet(fitpsf_file_name)
        >>> image_pos = galsim.PositionD(image_x, image_y)    # position in pixels on the image
        >>>                                                   # NOT in arcsec on the sky!
        >>> psf = des_shapelet.getPSF(image_pos)   # profile is in world coordinates

    Note that the returned psf here already includes the pixel.  This is what is sometimes
    called an "effective PSF".  Thus, you should not convolve by the pixel profile again
    (nor integrate over the pixel).  This would effectively include the pixel twice!

    This class will only interpolate within the defining bounds.  It won't extrapolate
    beyond the bounding box of where the stars defined the interpolation.
    If you try to use it with an invalid position, it will throw an IndexError.
    You can check whether a position is valid with

        >>> if des_shapelet.bounds.includes(pos):
        >>>     psf = des_shapelet.getPSF(pos)
        >>> else:
        >>>     [...skip this object...]


    Parameters:
        file_name:  The name of the file to be read in.
        dir:        Optionally a directory name can be provided if the file names do not
                    already include it. [default: None]
    """
    _req_params = { 'file_name' : str }
    _opt_params = { 'dir' : str }
    _single_params = []
    _takes_rng = False

    def __init__(self, file_name, dir=None):
        if dir:
            file_name = os.path.join(dir,file_name)
        self.file_name = file_name
        self.read_fits()

    def read_fits(self):
        """Read in a DES_Shapelet stored in FITS file.
        """
        with pyfits.open(self.file_name) as fits:
            cat = fits[1].data
        # These fields each only contain one element, hence the [0]'s.
        self.psf_order = cat.field('psf_order')[0]
        self.psf_size = (self.psf_order+1) * (self.psf_order+2) // 2
        self.sigma = cat.field('sigma')[0]
        self.fit_order = cat.field('fit_order')[0]
        self.fit_size = (self.fit_order+1) * (self.fit_order+2) // 2
        self.npca = cat.field('npca')[0]

        self.bounds = BoundsD(
            float(cat.field('xmin')[0]), float(cat.field('xmax')[0]),
            float(cat.field('ymin')[0]), float(cat.field('ymax')[0]))

        self.ave_psf = cat.field('ave_psf')[0]
        assert self.ave_psf.shape == (self.psf_size,)

        # Note: older pyfits versions don't get the shape right.
        # For newer pyfits versions the reshape command should be a no op.
        self.rot_matrix = cat.field('rot_matrix')[0].reshape((self.psf_size,self.npca)).T
        assert self.rot_matrix.shape == (self.npca, self.psf_size)

        self.interp_matrix = cat.field('interp_matrix')[0].reshape((self.npca,self.fit_size)).T
        assert self.interp_matrix.shape == (self.fit_size, self.npca)

    def getBounds(self):
        return self.bounds

    def getSigma(self):
        return self.sigma

    def getOrder(self):
        return self.psf_order

    def getPSF(self, image_pos, gsparams=None):
        """Returns the PSF at position image_pos

        Parameters:
            image_pos:  The position in pixel units for which to build the PSF.
            gsparams:   An optional `GSParams` instance to pass to the constructed GSObject.
                        [default: None]

        Returns:
            the PSF as a galsim.Shapelet instance
        """
        psf = Shapelet(self.sigma, self.psf_order, self.getB(image_pos), gsparams=gsparams)

        # The fitpsf files were built with respect to (u,v) = (ra,dec).  The GalSim convention is
        # to use sky coordinates with u = -ra.  So we need to flip the profile across the v axis
        # to take u -> -u.
        psf = psf.transform(-1,0,0,1)

        return psf

    def getB(self, pos):
        """Get the B vector as a numpy array at position pos
        """
        if not self.bounds.includes(pos):
            raise GalSimBoundsError("position in DES_Shapelet.getPSF is out of bounds",
                                    pos, self.bounds)

        Px = self._definePxy(pos.x,self.bounds.xmin,self.bounds.xmax)
        Py = self._definePxy(pos.y,self.bounds.ymin,self.bounds.ymax)
        order = self.fit_order
        P = np.array([ Px[n-q] * Py[q] for n in range(order+1) for q in range(n+1) ])
        assert len(P) == self.fit_size

        # Note: This is equivalent to:
        #
        #     P = numpy.empty(self.fit_size)
        #     k = 0
        #     for n in range(self.fit_order+1):
        #         for q in range(n+1):
        #             P[k] = Px[n-q] * Py[q]
        #             k = k+1

        b1 = np.dot(P,self.interp_matrix)
        b = np.dot(b1,self.rot_matrix)
        assert len(b) == self.psf_size
        b += self.ave_psf
        return b

    def _definePxy(self, x, min, max):
        x1 = (2.*x-min-max)/(max-min)
        temp = np.empty(self.fit_order+1)
        temp[0] = 1
        if self.fit_order > 0:  # pragma: no branch (always true for file we have for testing.)
            temp[1] = x1
        for i in range(2,self.fit_order+1):
            temp[i] = ((2.*i-1.)*x1*temp[i-1] - (i-1.)*temp[i-2]) / float(i)
        return temp

# First we need to add the class itself as a valid input_type.
RegisterInputType('des_shapelet', InputLoader(DES_Shapelet))

# Also make a builder to create the PSF object for a given position.
# The builders require 4 args.
# config is a dictionary that includes 'type' plus other items you might want to allow or require.
# base is the top level config dictionary where some global variables are stored.
# ignore is a list of key words that might be in the config dictionary that you should ignore.
def BuildDES_Shapelet(config, base, ignore, gsparams, logger):
    """Build a GSObject representing the shapelet model at the correct location in the image in a
    config-processing context.

    This is used as object type ``DES_Shapelet`` in a config file.

    It requires the use of the ``des_shapelet`` input field.
    """
    des_shapelet = GetInputObj('des_shapelet', config, base, 'DES_Shapelet')

    opt = { 'flux' : float , 'num' : int, 'image_pos' : PositionD }
    params, safe = GetAllParams(config, base, opt=opt, ignore=ignore)

    if 'image_pos' in params:
        image_pos = params['image_pos']
    elif 'image_pos' in base:
        image_pos = base['image_pos']
    else:
        raise GalSimConfigError("DES_Shapelet requested, but no image_pos defined in base.")

    # Convert gsparams from a dict to an actual GSParams object
    if gsparams: gsparams = GSParams(**gsparams)
    else: gsparams = None

    if des_shapelet.getBounds().includes(image_pos):
        #psf = des_shapelet.getPSF(image_pos, gsparams)
        # Because of serialization issues, the above call doesn't work.  So we need to
        # repeat the internals of getPSF here.
        b = des_shapelet.getB(image_pos)
        sigma = des_shapelet.getSigma()
        order = des_shapelet.getOrder()
        psf = Shapelet(sigma, order, b, gsparams=gsparams).transform(-1,0,0,1)
    else:
        message = 'Position '+str(image_pos)+' not in interpolation bounds: '
        message += str(des_shapelet.getBounds())
        raise SkipThisObject(message)

    if 'flux' in params:
        psf = psf.withFlux(params['flux'])

    # The second item here is "safe", a boolean that declares whether the returned value is
    # safe to save and use again for later objects.  In this case, we wouldn't want to do
    # that, since they will be at different positions, so the interpolated PSF will be different.
    return psf, False

# Register this builder with the config framework:
RegisterObjectType('DES_Shapelet', BuildDES_Shapelet, input_type='des_shapelet')

